#include "catmull.h"
#include "catmull_cut_bezier.h"
#include <iostream>
#include <assert.h>
#include <iostream>
#include <algorithm>
#include <vector>
using namespace std;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
double Catmull::pixelSize=1.0/512.0;
double Catmull::pas=0.1;
int Catmull::NB_STEPS=(int)(1.0/0.1);
int Catmull::lineWidth=1;
double Catmull::controlPointsSize=0.03;
GLSLProgram Catmull::colorConstraintsProgram;
int Catmull::wRef = 512;
int Catmull::hRef = 512;
GLuint Catmull::refBuffer=0; 
int Catmull::lifetimeThreshold = 0;
QImage* Catmull::exampleImg=NULL;
QImage* Catmull::exampleGradient=NULL;


void Catmull::setUpDefaultParam()
{
    nbPoint=0;
    controlPoints.reserve(4);
    leftColors.reserve(4);
    rightColors.reserve(4);
    movingPoint = -1;
    lifetime = MAX_LIFETIME;

    selectedSpline = false;
    closedCurve = false;
    bDrawStraight = false;
}

Catmull::Catmull()
{
    leftColor = QColor(leftColorInit[0], leftColorInit[1], leftColorInit[2]);
    rightColor = QColor(rightColorInit[0], rightColorInit[1], rightColorInit[2]);

    sigma=0;
    setUpDefaultParam();
}

Catmull::Catmull(QColor lC, QColor rC, int s)
{
    leftColor=lC;
    rightColor=rC;

    sigma=s;
    setUpDefaultParam();
}

Catmull::~Catmull()
{

}

//interpolate between curves c1 and c2: cResult = t*c2 + (1-t)*c1, 0<=t<=1
//precondition: c1 and c2 have the same number of control points
void Catmull::interpolate(Catmull* c1, Catmull* c2, Catmull* cResult, double t)
{
    cResult->controlPoints.clear();
    cResult->leftColors.clear();
    cResult->rightColors.clear();
    cResult->sigmas.clear();

    cResult->nbPoint = c1->nbPoint;
    cResult->nbLeftColors = c1->nbLeftColors; 
    cResult->nbRightColors = c1->nbRightColors;
    cResult->nbSigmas = c1->nbSigmas;
    cResult->leftColorsIDs = c1->leftColorsIDs;
    cResult->rightColorsIDs = c1->rightColorsIDs;
    cResult->sigmasIDs = c1->sigmasIDs;
    cResult->lifetime = (int)((1.0-t)*c1->lifetime+t*c2->lifetime+0.5);

	//positions
    cResult->controlPoints.reserve(cResult->nbPoint);
    for(int i=0;i<cResult->nbPoint;i++){
        Point P((1.0-t)*c1->controlPoints[i].getX()+t*c2->controlPoints[i].getX(),
                 (1.0-t)*c1->controlPoints[i].getY()+t*c2->controlPoints[i].getY());
        cResult->controlPoints.insert(cResult->controlPoints.end(),P);
}
	//colors
    cResult->leftColors.reserve(cResult->nbLeftColors);
    for(int i=0;i<cResult->nbLeftColors;i++){
        QColor C((int)((1.0-t)*(double)c1->leftColors[i].red()+t*(double)c2->leftColors[i].red()+0.5),
                  (int)((1.0-t)*(double)c1->leftColors[i].green()+t*(double)c2->leftColors[i].green()+0.5),
                  (int)((1.0-t)*(double)c1->leftColors[i].blue()+t*(double)c2->leftColors[i].blue())+0.5);
        cResult->leftColors.insert(cResult->leftColors.end(),C);
}
	
    cResult->rightColors.reserve(cResult->nbRightColors);
    for(int i=0;i<cResult->nbRightColors;i++){
        QColor C((int)((1.0-t)*(double)c1->rightColors[i].red()+t*(double)c2->rightColors[i].red()+0.5),
                  (int)((1.0-t)*(double)c1->rightColors[i].green()+t*(double)c2->rightColors[i].green()+0.5),
                  (int)((1.0-t)*(double)c1->rightColors[i].blue()+t*(double)c2->rightColors[i].blue())+0.5);
        cResult->rightColors.insert(cResult->rightColors.end(),C);
}
	//blur
    cResult->sigmas.reserve(cResult->nbSigmas);
    for(int i=0;i<cResult->nbSigmas;i++){
        int s = (int)((1.0-t)*(double)c1->sigmas[i]+t*(double)c2->sigmas[i]+0.5);
        cResult->sigmas.insert(cResult->sigmas.end(),s);
}
}

//the curve is closed or not
void Catmull::setClosedCurve(bool _closedCurve)
{
    if(nbPoint<2) return;

    closedCurve = _closedCurve;
    if(closedCurve)
{
        int nbDecalSteps = 1;
		//set the same color
		//all curves have to have colors at beginning and end points
        if(leftColors[0]!=leftColors[nbLeftColors-1])
{
			//a spline has at least NB_STEPS interpolated points;
			//NB_STEPS = 10;
            QColor lC = leftColors[0];
            leftColors[0] = leftColors[nbLeftColors-1];
            leftColors.insert(1, lC);
            leftColorsIDs.insert(1, nbDecalSteps);
            nbLeftColors = leftColors.size();
}
        if(rightColors[0]!=rightColors[nbRightColors-1])
{
			//a spline has at least NB_STEPS interpolated points;
			//NB_STEPS = 10;
            QColor rC = rightColors[0];
            rightColors[0] = rightColors[nbRightColors-1];

            rightColors.insert(1, rC);
            rightColorsIDs.insert(1, nbDecalSteps);
            nbRightColors = rightColors.size();
}
        if(sigmas[0]!=sigmas[nbSigmas-1])
{
			//a spline has at least NB_STEPS interpolated points;
			//NB_STEPS = 10;
            int s = sigmas[0];
            sigmas[0] = sigmas[nbSigmas-1];
            sigmas.insert(1, s);
            sigmasIDs.insert(1, nbDecalSteps);
            nbSigmas = sigmas.size();
}
}
}
bool Catmull::isClosedCurve()
{
    return closedCurve;
}

//init the pixelSize, reference size, reference buffer id and shaders. Must be called once OpenGl is all set.
void Catmull::initStaticDatas(double pSize, int wR, int hR, GLuint idR, QImage* img, QImage* grad)
{
    if (!colorConstraintsProgram.create("shaders/color_constraints.vert","shaders/color_constraints.frag"))
{
        cerr<<"Color constraints Shader : "<<colorConstraintsProgram.lastError()<<endl;
}
    pixelSize=pSize;
    wRef = wR;
    hRef = hR;
    refBuffer = idR;
    exampleImg = img;
    exampleGradient = grad;
}

void Catmull::zoomIn()
{
    controlPointsSize/=2.0;
}

void Catmull::zoomOut()
{
    controlPointsSize*=2.0;
}

void Catmull::setLifetimeThreshold(int th)
{
    lifetimeThreshold = th;
}

//******************
//update the IDs whn a geometry control point is deleted or inserted (so the total nb of steps changes)
void Catmull::updateAttribIDs(int controlPointINDEX, bool isBeingAdded)
{
	//if out of bounds, return;
    if(controlPointINDEX < 0 || controlPointINDEX > nbPoint-1)
        return;

    int globalID = (controlPointINDEX/3)*NB_STEPS;
	//if end points, special case
    int leftColorINDEX = 0;
    int rightColorINDEX = 0;
    int sigmaINDEX = 0;
    double leftColorT;
    double rightColorT;
    double sigmaT;
    QColor Cl, Cr;
    double blur;


    if(controlPointINDEX == 0)
{
		//the next destination point (P3) will become the origin.
		//its attributes are aded to the attribute control points 
		//all attributes that influence the spline P0-P3 are deleted.

		//compute colors
        leftColorT = ComputeAttribT(1, controlPointINDEX, leftColorINDEX, COLOR_LEFT);
        Cl=ComputeCLeft(leftColorT,leftColorINDEX);
        rightColorT = ComputeAttribT(1, controlPointINDEX, rightColorINDEX, COLOR_RIGHT);
        Cr=ComputeCRight(rightColorT, rightColorINDEX);
		//compute the blur
        sigmaT = ComputeAttribT(1, controlPointINDEX, sigmaINDEX, BLUR);
        blur = ComputeBlur(sigmaT, sigmaINDEX);

        bool passedFirstSpline = false;
        while(leftColors.size()>1 && !passedFirstSpline)
{
            if(leftColorsIDs[0]<NB_STEPS)
{
                leftColorsIDs.pop_front();
                leftColors.pop_front();
}
            else passedFirstSpline = true;
}
        leftColors.insert(leftColors.begin(), Cl);
		//it will be decreased to 0 the next step
        leftColorsIDs.insert(leftColorsIDs.begin(), NB_STEPS);
        nbLeftColors = leftColors.size();

        passedFirstSpline = false;
        while(rightColors.size()>1 && !passedFirstSpline)
{
            if(rightColorsIDs[0]<NB_STEPS)
{
                rightColorsIDs.pop_front();
                rightColors.pop_front();
}
            else passedFirstSpline = true;
}
        rightColors.insert(rightColors.begin(), Cr);
        rightColorsIDs.insert(rightColorsIDs.begin(), NB_STEPS);
        nbRightColors = rightColors.size();

        passedFirstSpline = false;
        while(sigmas.size()>1 && !passedFirstSpline)
{
            if(sigmasIDs[0]<NB_STEPS)
{
                sigmasIDs.pop_front();
                sigmas.pop_front();
}
            else passedFirstSpline = true;
}
        sigmas.insert(sigmas.begin(), blur);
        sigmasIDs.insert(sigmasIDs.begin(), NB_STEPS);
        nbSigmas = sigmas.size();
}

    if(controlPointINDEX == nbPoint-1)
{
		//the next origin point (P(n-4)) will become the "final destination".
		//its attributes are aded to the attribute control points 
		//all attributes that influence the spline P(n-4)-P(n-1) are deleted.

		//compute colors
        leftColorT = ComputeAttribT(0, controlPointINDEX-3, leftColorINDEX, COLOR_LEFT);
        Cl=ComputeCLeft(leftColorT,leftColorINDEX);
        rightColorT = ComputeAttribT(0, controlPointINDEX-3, rightColorINDEX, COLOR_RIGHT);
        Cr=ComputeCRight(rightColorT, rightColorINDEX);
		//compute the blur
        sigmaT = ComputeAttribT(0, controlPointINDEX-3, sigmaINDEX, BLUR);
        blur = ComputeBlur(sigmaT, sigmaINDEX);

        bool passedFirstSpline = false;
        while(leftColors.size()>1 && !passedFirstSpline)
{
            if(leftColorsIDs.last()>globalID-NB_STEPS)
{
                leftColorsIDs.pop_back();
                leftColors.pop_back();
}
            else passedFirstSpline = true;
}
        leftColors.push_back(Cl);
        leftColorsIDs.push_back(globalID);
        nbLeftColors = leftColors.size();

        passedFirstSpline = false;
        while(rightColors.size()>1 && !passedFirstSpline)
{
            if(rightColorsIDs.last()>globalID-NB_STEPS)
{
                rightColorsIDs.pop_back();
                rightColors.pop_back();
}
            else passedFirstSpline = true;
}
        rightColors.push_back(Cr);
        rightColorsIDs.push_back(globalID);
        nbRightColors = rightColors.size();

        passedFirstSpline = false;
        while(sigmas.size()>1 && !passedFirstSpline)
{
            if(sigmasIDs.last()>globalID-NB_STEPS)
{
                sigmasIDs.pop_back();
                sigmas.pop_back();
}
            else passedFirstSpline = true;
}
        sigmas.push_back(blur);
        sigmasIDs.push_back(globalID);
        nbSigmas = sigmas.size();
}

	//proceed to update attribute IDs
    for(int i=0;i<nbLeftColors;i++)
{
        if(leftColorsIDs[i]>=globalID) 
            if(isBeingAdded)
                leftColorsIDs[i]+=NB_STEPS;
        else
            leftColorsIDs[i]-=NB_STEPS;
}

    for(int i=0;i<nbRightColors;i++)
        if(rightColorsIDs[i]>=globalID) 
            if(isBeingAdded)
                rightColorsIDs[i]+=NB_STEPS;
    else
        rightColorsIDs[i]-=NB_STEPS;

    for(int i=0;i<nbSigmas;i++)
        if(sigmasIDs[i]>=globalID) 
            if(isBeingAdded)
                sigmasIDs[i]+=NB_STEPS;
    else
        sigmasIDs[i]-=NB_STEPS;
}

//returns a value between 0 and 1, and fills in geomINDEX
double Catmull::ComputeGeomT(int globalID, int& geomINDEX)
{
    geomINDEX = (globalID/NB_STEPS)*3;
    double t = (double)(globalID%NB_STEPS)/(double)NB_STEPS;
    return t;
}

//returns a value between 0 and 1, used to interpolate color between color control points
//and update the attribute index if necessary
double Catmull::ComputeAttribT(double geomT, int geomINDEX, int& attribINDEX, AttribType attribTYPE)
{
    int geomID = (geomINDEX/3)*NB_STEPS;
    double globalID = (geomT*NB_STEPS);
	//modf ((geomT*NB_STEPS), &globalID);
    globalID += geomID;//global t, from 0 to NB_STEPS*nbBeziers

    int nbAttribs;
    QVector<int>* attribIDs;
    switch(attribTYPE)
{
    case COLOR_LEFT:
            nbAttribs = nbLeftColors;
            attribIDs = &leftColorsIDs;
            break;
    case COLOR_RIGHT:
            nbAttribs = nbRightColors;
            attribIDs = &rightColorsIDs;
            break;
    case BLUR:
            nbAttribs = nbSigmas;
            attribIDs = &sigmasIDs;
            break;
    default:
            nbAttribs = nbSigmas;
            attribIDs = &sigmasIDs;
            break;
}

    assert(attribINDEX<nbAttribs-1);
    if(globalID>= attribIDs->at(attribINDEX+1))
{
        attribINDEX++;
        return 0;
}
    double pasColor = 1./(double)(attribIDs->at(attribINDEX+1) - attribIDs->at(attribINDEX));
    return (globalID-attribIDs->at(attribINDEX))*pasColor;
}


//returns a value between 0 and 1, used to interpolate color between color control points
//and compute the attribute index
double Catmull::ComputeAttribTAttribIndex(double geomT, int geomINDEX, int& attribINDEX, AttribType attribTYPE)
{
    int geomID = (geomINDEX/3)*NB_STEPS;
    double globalID = (geomT*NB_STEPS);
	//modf ((geomT*NB_STEPS), &globalID);
    globalID += geomID;//global t, from 0 to NB_STEPS*nbBeziers

    int nbAttribs;
    QVector<int>* attribIDs;
    switch(attribTYPE)
{
    case COLOR_LEFT:
            nbAttribs = nbLeftColors;
            attribIDs = &leftColorsIDs;
            break;
    case COLOR_RIGHT:
            nbAttribs = nbRightColors;
            attribIDs = &rightColorsIDs;
            break;
    case BLUR:
            nbAttribs = nbSigmas;
            attribIDs = &sigmasIDs;
            break;
    default:
            nbAttribs = nbSigmas;
            attribIDs = &sigmasIDs;
            break;
}

    int id=nbAttribs-1;
        //globalID is the position of the geometric point in the "global view" of all splines
        //looking for the last attribute that has a value less than globalID
    while(id>=0 && globalID>attribIDs->at(id)){
        id--;
}
    attribINDEX = id;
    if(attribINDEX==attribIDs->size()-1) return 0;
    double pasColor = 1./(double)(attribIDs->at(attribINDEX+1) - attribIDs->at(attribINDEX));
    return (globalID-attribIDs->at(attribINDEX))*pasColor;
}

//compute gradient amplitude G(t)
double Catmull::ComputeGr(double left_red, double right_red)
{
    double Gr;
    Gr=(left_red-right_red)/255.f;

    return Gr;
}

//compute gradient amplitude G(t)
double Catmull::ComputeGg(double left_green, double right_green)
{
    double Gg;
    Gg = (left_green-right_green)/255.f;
    return Gg;
}

//compute gradient amplitude G(t)
double Catmull::ComputeGb(double left_blue, double right_blue)
{
    double Gb;
    Gb = (left_blue-right_blue)/255.f;
    return Gb;
}

//compute left color Cleft(t)
QColor Catmull::ComputeCLeft(double t, int start)
{
    QColor Cl;
    if(start == nbLeftColors-1)
        return leftColors[start];
		
    Cl.setRed((int)((1-t)*(double)leftColors[start].red()+t*(double)leftColors[start+1].red())+0.5);
    Cl.setGreen((int)((1-t)*(double)leftColors[start].green()+t*(double)leftColors[start+1].green())+0.5);
    Cl.setBlue((int)((1-t)*(double)leftColors[start].blue()+t*(double)leftColors[start+1].blue())+0.5);
    return Cl;
}

//compute right color Cright(t)
QColor Catmull::ComputeCRight(double t, int start)
{
    QColor Cr;
    if(start == nbRightColors-1)
        return rightColors[start];
    Cr.setRed((int)((1-t)*(double)rightColors[start].red()+t*(double)rightColors[start+1].red()));
    Cr.setGreen((int)((1-t)*(double)rightColors[start].green()+t*(double)rightColors[start+1].green()));
    Cr.setBlue((int)((1-t)*(double)rightColors[start].blue()+t*(double)rightColors[start+1].blue()));

    return Cr;
}

//compute the blur sigma(t)
int Catmull::ComputeBlur(double t, int start)
{
    int s;
    if(start == nbSigmas-1)
        return sigmas[start];
		
    s=(int)((1.0-t)*(double)sigmas[start]+t*(double)sigmas[start+1]+0.5);

    return s;
}

//compute P(t)
/*
* Evaluate a Bezier curve at parameter value t.
*/
Point Catmull::ComputeP(double tVal, int startPoint)
{
    int start; double t;
    if(startPoint == nbPoint-1)
{
        t = 1-tVal;
        start = startPoint-3;
}
    else
{
        t = tVal;
        start = startPoint;
}
    assert(start<nbPoint-3);

    Point P;
    double x,y;
    double a, b, c;
    Point P0, P1, P2, P3;
    P0 = controlPoints[start];
    P1 = controlPoints[start+1];
    P2 = controlPoints[start+2];
    P3 = controlPoints[start+3];
    
        //if stylization: linear interpolation
    if(bDrawStraight){
        P.setX((1.0-t)*P0.getX()+t*P3.getX());
        P.setY((1.0-t)*P0.getY()+t*P3.getY());
}
    else{//Bezier
        c = 3*(P1.getX() - P0.getX());
        b = 3*(P2.getX() - P1.getX()) - c;
        a = P3.getX() - P0.getX() - c - b; 
        x = a*t*t*t + b*t*t + c*t + P0.getX();
    
        c = 3*(P1.getY() - P0.getY());
        b = 3*(P2.getY() - P1.getY()) - c;
        a = P3.getY() - P0.getY() - c - b; 
        y = a*t*t*t + b*t*t + c*t + P0.getY();
    
        P.setX(x);
        P.setY(y);
}
    return P;

    //--- Catmull-Rom ---//	
	//Point P;

	//double x,y;

	//x = 0.5*(2.0*controlPoints[start+1].getX()
	//	+(-controlPoints[start].getX()+controlPoints[start+2].getX())*t
	//	+(2.0*controlPoints[start].getX()-5.0*controlPoints[start+1].getX()
	//	+4.0*controlPoints[start+2].getX()-controlPoints[start+3].getX())*t*t
	//	+(-controlPoints[start].getX()+3.0*controlPoints[start+1].getX()
	//	-3.0*controlPoints[start+2].getX()+controlPoints[start+3].getX())*t*t*t);
	//y = 0.5*(2.0*controlPoints[start+1].getY()
	//	+(-controlPoints[start].getY()+controlPoints[start+2].getY())*t
	//	+(2.0*controlPoints[start].getY()-5.0*controlPoints[start+1].getY()
	//	+4.0*controlPoints[start+2].getY()-controlPoints[start+3].getY())*t*t
	//	+(-controlPoints[start].getY()+3.0*controlPoints[start+1].getY()
	//	-3.0*controlPoints[start+2].getY()+controlPoints[start+3].getY())*t*t*t);

	//P.setX(x);
	//P.setY(y);
	//return P;
}

//compute P'(t)
Point Catmull::ComputePNormal(double tVal, int startPoint)
{
    int start; double t;
    if(startPoint == nbPoint-1)
    {
        t = 1-tVal;
        start = startPoint-3;
    }
    else
    {
        t = tVal;
        start = startPoint;
    }
    assert(start<nbPoint-3);
    Point P;

    double x,y;
    double a, b, c;
    Point P0, P1, P2, P3;
    P0 = controlPoints[start];
    P1 = controlPoints[start+1];
    P2 = controlPoints[start+2];
    P3 = controlPoints[start+3];

        //if stylization: straight line
    if(bDrawStraight){
        x = P3.getX()-P0.getX();
        y = P3.getY()-P0.getY();
    }
    else{
            // Bezier derivative
        c = 3*(P1.getX() - P0.getX());
        b = 3*(P2.getX() - P1.getX()) - c;
        a = P3.getX() - P0.getX() - c - b; 
        x = 3*a*t*t + 2*b*t + c;
    
        c = 3*(P1.getY() - P0.getY());
        b = 3*(P2.getY() - P1.getY()) - c;
        a = P3.getY() - P0.getY() - c - b; 
        y = 3*a*t*t + 2*b*t + c;
    }
    double norm = sqrt(x*x+y*y);
    x /= norm;
    y /= norm;
    P.setX(-y);
    P.setY(x);
    return P;

	
    //--- Catmull-Rom ---//
	//Point P;

	//double x,y;

	////derivative
	//x = 0.5*(-controlPoints[start].getX()+controlPoints[start+2].getX()
	//	+2.0*(2.0*controlPoints[start].getX()-5.0*controlPoints[start+1].getX()
	//	+4.0*controlPoints[start+2].getX()-controlPoints[start+3].getX())*t
	//	+3.0*(-controlPoints[start].getX()+3.0*controlPoints[start+1].getX()
	//	-3.0*controlPoints[start+2].getX()+controlPoints[start+3].getX())*t*t);
	//y = 0.5*(-controlPoints[start].getY()+controlPoints[start+2].getY()
	//	+2.0*(2.0*controlPoints[start].getY()-5.0*controlPoints[start+1].getY()
	//	+4.0*controlPoints[start+2].getY()-controlPoints[start+3].getY())*t
	//	+3.0*(-controlPoints[start].getY()+3.0*controlPoints[start+1].getY()
	//	-3.0*controlPoints[start+2].getY()+controlPoints[start+3].getY())*t*t);

	//double norm = sqrt(x*x+y*y);
	//x /= norm;
	//y /= norm;
	//P.setX(-y);
	//P.setY(x);
	//return P;
}

//compute tangent at P(t)
Point Catmull::ComputePTangent(double tVal, int startPoint)
{
    int start; double t;
    if(startPoint == nbPoint-1)
    {
        t = 1-tVal;
        start = startPoint-3;
    }
    else
    {
        t = tVal;
        start = startPoint;
    }
    assert(start<nbPoint-3);
    Point P;

    double x,y;
    double a, b, c;
    Point P0, P1, P2, P3;
    P0 = controlPoints[start];
    P1 = controlPoints[start+1];
    P2 = controlPoints[start+2];
    P3 = controlPoints[start+3];

        //if stylization: straight line
    if(bDrawStraight){
        x = P3.getX()-P0.getX();
        y = P3.getY()-P0.getY();
    }
    else{
            // Bezier derivative
        c = 3*(P1.getX() - P0.getX());
        b = 3*(P2.getX() - P1.getX()) - c;
        a = P3.getX() - P0.getX() - c - b; 
        x = 3*a*t*t + 2*b*t + c;
    
        c = 3*(P1.getY() - P0.getY());
        b = 3*(P2.getY() - P1.getY()) - c;
        a = P3.getY() - P0.getY() - c - b; 
        y = 3*a*t*t + 2*b*t + c;
    }
    P.setX(x);
    P.setY(y);
    return P;
}

//add the (x,y) point in the control point list
void Catmull::addPoint(double x,double y)
{   
	//position
    Point P(x,y);
    controlPoints.insert(controlPoints.end(),P);
    nbPoint++;
    updateBoundingBox();
}

void Catmull::addPolLinePoint(double x, double y)
{
	Point P(x,y);
    polLinePoints.insert(polLinePoints.end(),P);
}

//**********************************//
//void Catmull::addLeftColorPoint(QColor lC)
//{
//	leftColors.insert(leftColors.end(),lC);
//}
//void Catmull::addRightColorPoint(QColor rC)
//{
//	rightColors.insert(rightColors.end(), rC);
//}

//******************************************//
//use the default color
void Catmull::addLeftColorPoint(int colorID, bool useDefault, QColor lC)
{
	//color
    QColor cl = useDefault? leftColor : lC;
    leftColors.insert(leftColors.end(),cl);
    leftColorsIDs.insert(leftColorsIDs.end(), colorID);
    nbLeftColors = leftColors.size();
}

void Catmull::addRightColorPoint(int colorID,  bool useDefault, QColor rC)
{
	//color
    QColor cR = useDefault? rightColor : rC;
    rightColors.insert(rightColors.end(),cR);
    rightColorsIDs.insert(rightColorsIDs.end(), colorID);
    nbRightColors = rightColors.size();
}

void Catmull::addBlurPoint(int blurID, bool useDefault, double sigmaBlur)//use the default blur
{
    double s = useDefault ? sigma:sigmaBlur;
    sigmas.insert(sigmas.end(),s);
    sigmasIDs.insert(sigmasIDs.end(),blurID);
    nbSigmas = sigmas.size();
}
//******************************************//
//---- deprecated ----//
void Catmull::addLeftColorPoint(bool refL, double x,double y)
{
	//color
    QColor cl = leftColor;
    if(refL){//if we have to read one of the colors in the reference buffer
        x = 0.5*(x+1);
        y = 0.5*(y+1);
        float color[4];
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,refBuffer);
        glReadPixels((GLint)(x*wRef),(GLint)(y*hRef),1,1,GL_RGBA,GL_FLOAT,color);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
        cl=QColor((int)(color[0]*255.0+0.5), (int)(color[1]*255.0+0.5), (int)(color[2]*255.0+0.5));
}

    leftColors.insert(leftColors.end(),cl);
}
void Catmull::addRightColorPoint(bool refR,double x,double y )
{
	//color
    QColor cr = rightColor;
    if(refR){//if we have to read one of the colors in the reference buffer
        x = 0.5*(x+1);
        y = 0.5*(y+1);
        float color[4];
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,refBuffer);
        glReadPixels((GLint)(x*wRef),(GLint)(y*hRef),1,1,GL_RGBA,GL_FLOAT,color);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
        cr=QColor((int)(color[0]*255.0+0.5), (int)(color[1]*255.0+0.5), (int)(color[2]*255.0+0.5));
}
    rightColors.insert(rightColors.end(),cr);
}
//-----//
//******************************************//

//set one side color by sampling colors in the reference image (attribTYPE specify wich side)
void Catmull::sampleColors(AttribType attribTYPE){
	int side;
	QVector<QColor>* colors;
	QVector<int>* colorsIDs;
	int* nbC;
	if(attribTYPE == COLOR_LEFT){
		side = 1;
		colors = &leftColors;
		colorsIDs = &leftColorsIDs;
		nbC = &nbLeftColors;
	}
	else{
		side = -1;
		colors = &rightColors;
		colorsIDs = &rightColorsIDs;
		nbC = &nbRightColors;
	}

	//sampling
	float color[4];
	float x,y;
	double offset=5.0*pixelSize;
	double t;
	QColor c;
	Point P,Pnormal;
	QVector<QColor> samples(0);
	QVector<int> samplesIDs(0);
	QVector<int> samplesPolLine(0);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,refBuffer);
	for(int i=0; i<nbPoint-2; i+=3){
		for(t=0;t<=1;t+=pas){
			P=ComputeP(t,i);//compute P(t)
			Pnormal=ComputePNormal(t,i);//compute the normal at P(t)

			x = 0.5*(P.getX()+side*offset*Pnormal.getX()+1);
			y = 0.5*(P.getY()+side*offset*Pnormal.getY()+1);

			glReadPixels((GLint)(x*wRef),(GLint)(y*hRef),1,1,GL_RGBA,GL_FLOAT,color);
			c=QColor((int)(color[0]*255.0+0.5), (int)(color[1]*255.0+0.5), (int)(color[2]*255.0+0.5));

			int id = ((i/3+t)*NB_STEPS+0.5);
			if(samplesIDs.size()==0 || samplesIDs.last()!=id){
				samples.insert(samples.end(),c);
				samplesIDs.insert(samplesIDs.end(),id);
			}	
		}
	}
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);

	//fit the samples, and put the colors in the color control points vector
	//last variable: the maximum Lab distance accepted betwwen colors
	samplesPolLine = vectorizedColorSet(samples, samplesIDs, 30);
	colors->clear();
	colorsIDs->clear();
	for(int i=0;i<samplesPolLine.size();i++)
	{
		colors->insert(colors->end(),samples[samplesPolLine[i]]);
		colorsIDs->insert(colorsIDs->end(), samplesIDs[samplesPolLine[i]]);    
		*nbC = colors->size();
	}
}
//sample the colors in the example image
void Catmull::sampleFromExample(){
	    
    //sampling
    float color[4];
    float x,y;
    double offset=6.0*pixelSize;
    double t;
    QColor c;
    Point P,Pnormal;
	leftColors.clear();
    leftColorsIDs.clear();
	rightColors.clear();
    rightColorsIDs.clear();
    for(int i=0; i<nbPoint-2; i+=3){
        for(t=0;t<=1;t+=pas){
            P=ComputeP(t,i);//compute P(t)
            Pnormal=ComputePNormal(t,i);//compute the normal at P(t)
            
			//left color
            x = 0.5*(P.getX()+offset*Pnormal.getX()+1);
            y = 1-0.5*(P.getY()+offset*Pnormal.getY()+1);
			if(x>=0 && x<1 && y>=0 && y<1){
				c=QColor(exampleImg->pixel(x*exampleImg->width(),y*exampleImg->height()));
                                int id = ((i/3+t)*NB_STEPS+0.5);
                                if(leftColorsIDs.size()==0 || leftColorsIDs.last()!=id){
                                    leftColors.insert(leftColors.end(),c);
                                    leftColorsIDs.insert(leftColorsIDs.end(),id);
                                }
			}
			
			//right color
            x = 0.5*(P.getX()-offset*Pnormal.getX()+1);
            y = 1-0.5*(P.getY()-offset*Pnormal.getY()+1);
			if(x>=0 && x<1 && y>=0 && y<1){
				c=QColor(exampleImg->pixel(x*exampleImg->width(),y*exampleImg->height()));
                                int id = ((i/3+t)*NB_STEPS+0.5);
                                if(rightColorsIDs.size()==0 || rightColorsIDs.last()!=id){
				    rightColors.insert(rightColors.end(),c);
                                    rightColorsIDs.insert(rightColorsIDs.end(),id);
                                }
			}
        }
    }
	

    
    //fit the samples, and put the colors in the color control points vector
	//last variable: the maximum Lab distance accepted betwwen colors
	QVector<int> samplesPolLine;
	QVector<QColor> samples;
	QVector<int> samplesIDs;
	samplesPolLine = vectorizedColorSet(leftColors, leftColorsIDs, 30);
	samples = leftColors;
	samplesIDs = leftColorsIDs;
	leftColors.clear();
	leftColorsIDs.clear();
	for(int i=0;i<samplesPolLine.size();i++)
	{
		leftColors.insert(leftColors.end(),samples[samplesPolLine[i]]);
		leftColorsIDs.insert(leftColorsIDs.end(), samplesIDs[samplesPolLine[i]]);    
	}

	nbLeftColors = leftColors.size();

	samplesPolLine.clear();
	samples.clear();
	samplesIDs.clear();

	samplesPolLine = vectorizedColorSet(rightColors, rightColorsIDs, 30);
	samples = rightColors;
	samplesIDs = rightColorsIDs;
	rightColors.clear();
	rightColorsIDs.clear();
	for(int i=0;i<samplesPolLine.size();i++)
	{
		rightColors.insert(rightColors.end(),samples[samplesPolLine[i]]);
		rightColorsIDs.insert(rightColorsIDs.end(), samplesIDs[samplesPolLine[i]]);    
	}

	nbRightColors = rightColors.size();
}

//insert a control point at a position given by globalID in the attribType control points list
void Catmull::insertAttribPoint(int globalID, AttribType attribTYPE)
{
    int nbAttribs; int attribINDEX;
    QVector<int>* attribIDs;
    int oldINDEX;
    double attribT;
    double geomT; int geomINDEX;
    double sigmaVal;
    QColor leftColorVal, rightColorVal;

    switch(attribTYPE)
{
    case COLOR_LEFT:
            nbAttribs = nbLeftColors;
            attribIDs = &leftColorsIDs;
            break;
    case COLOR_RIGHT:
            nbAttribs = nbRightColors;
            attribIDs = &rightColorsIDs;
            break;
    case BLUR:
            nbAttribs = nbSigmas;
            attribIDs = &sigmasIDs;
            break;
}

	//find the index of the first attribute that has the global ID bigger than the one we want to insert
	//(it's the attribute point positioned right after the one we are inserting)
    attribINDEX = 0;
    while ( (attribINDEX<nbAttribs-1) && (globalID>= attribIDs->at(attribINDEX)) )
{
        attribINDEX++;
}
    attribINDEX = min(attribINDEX, nbAttribs-1);
	//find the value currenlty used at the position we clicked
    oldINDEX = attribINDEX-1;
	//returns a value between 0 and 1, and fills in geomINDEX
    geomT = ComputeGeomT(globalID, geomINDEX);
	//returns a value between 0 and 1, used to interpolate color between color control points
    attribT = ComputeAttribT(geomT, geomINDEX, oldINDEX, attribTYPE);

    switch(attribTYPE)
{
    case COLOR_LEFT:
            leftColorVal = ComputeCLeft(attribT, oldINDEX);
            leftColorsIDs.insert(attribINDEX, globalID);
            leftColors.insert(attribINDEX, leftColorVal);
            nbLeftColors = leftColors.size();
            break;
    case COLOR_RIGHT:
            rightColorVal = ComputeCRight(attribT, oldINDEX);
            rightColorsIDs.insert(attribINDEX, globalID);
            rightColors.insert(attribINDEX, rightColorVal);
            nbRightColors = rightColors.size();
            break;
    case BLUR:
            sigmaVal = ComputeBlur(attribT, oldINDEX);
            sigmasIDs.insert(attribINDEX, globalID);
            sigmas.insert(attribINDEX, sigmaVal);
            nbSigmas = sigmas.size();
            break;
}
}

//insert a control point at a position given by globalID in the attribType control points list
//and return its position in the attribType control points list
int Catmull::insertAttribPointPos(int globalID, AttribType attribTYPE)
{
    int nbAttribs; int attribINDEX;
    QVector<int>* attribIDs;
    int oldINDEX;
    double attribT;
    double geomT; int geomINDEX;
    double sigmaVal;
    QColor leftColorVal, rightColorVal;

    switch(attribTYPE)
{
    case COLOR_LEFT:
            nbAttribs = nbLeftColors;
            attribIDs = &leftColorsIDs;
            break;
    case COLOR_RIGHT:
            nbAttribs = nbRightColors;
            attribIDs = &rightColorsIDs;
            break;
    case BLUR:
            nbAttribs = nbSigmas;
            attribIDs = &sigmasIDs;
            break;
}

	//find the index of the first attribute that has the global ID bigger than the one we want to insert
	//(it's the attribute point positioned right after the one we are inserting)
    attribINDEX = 0;
    while ( (attribINDEX<nbAttribs-1) && (globalID>= attribIDs->at(attribINDEX)) )
{
        attribINDEX++;
}
    attribINDEX = min(attribINDEX, nbAttribs-1);
	//find the value currenlty used at the position we clicked
    oldINDEX = attribINDEX-1;
	//returns a value between 0 and 1, and fills in geomINDEX
    geomT = ComputeGeomT(globalID, geomINDEX);
	//returns a value between 0 and 1, used to interpolate color between color control points
    attribT = ComputeAttribT(geomT, geomINDEX, oldINDEX, attribTYPE);

    switch(attribTYPE)
{
    case COLOR_LEFT:
            leftColorVal = ComputeCLeft(attribT, oldINDEX);
            leftColorsIDs.insert(attribINDEX, globalID);
            leftColors.insert(attribINDEX, leftColorVal);
            nbLeftColors = leftColors.size();
            break;
    case COLOR_RIGHT:
            rightColorVal = ComputeCRight(attribT, oldINDEX);
            rightColorsIDs.insert(attribINDEX, globalID);
            rightColors.insert(attribINDEX, rightColorVal);
            nbRightColors = rightColors.size();
            break;
    case BLUR:
            sigmaVal = ComputeBlur(attribT, oldINDEX);
            sigmasIDs.insert(attribINDEX, globalID);
            sigmas.insert(attribINDEX, sigmaVal);
            nbSigmas = sigmas.size();
            break;
}
    
    return attribINDEX;
}

//erase the control point closest to globalID in the attribType control points list
void Catmull::eraseAttribPoint(int globalID, AttribType attribTYPE) 
{
    int attribINDEX;
    switch(attribTYPE)
{
    case COLOR_LEFT:
		//find the index of the last attribute that has the global ID smaller than globalID
            attribINDEX = nbLeftColors-1;
            if(attribINDEX<2)return;
            while ( (attribINDEX>0) && (globalID< leftColorsIDs[attribINDEX] ) )
{
                attribINDEX--;
}
			if(attribINDEX>=nbLeftColors-1) return;
            leftColorsIDs.erase(leftColorsIDs.begin()+attribINDEX);
            leftColors.erase(leftColors.begin()+attribINDEX);
            nbLeftColors = leftColors.size();
            break;
    case COLOR_RIGHT:
		//find the index of the last attribute that has the global ID smaller than globalID
            attribINDEX = nbRightColors-1;
            if(attribINDEX<2)return;
            while ( (attribINDEX>0) && (globalID< rightColorsIDs[attribINDEX] ) )
{
                attribINDEX--;
}
			if(attribINDEX>=nbRightColors-1) return;
            rightColorsIDs.erase(rightColorsIDs.begin()+attribINDEX);
            rightColors.erase(rightColors.begin()+attribINDEX);
            nbRightColors = rightColors.size();
            break;
    case BLUR:
		//find the index of the last attribute that has the global ID smaller than globalID
            attribINDEX = nbSigmas-1;
            if(attribINDEX<2)return;
            while ( (attribINDEX>0) && (globalID< sigmasIDs[attribINDEX] ) )
{
                attribINDEX--;
}
			if(attribINDEX>=nbSigmas-1) return;
            sigmasIDs.erase(sigmasIDs.begin()+attribINDEX);
            sigmas.erase(sigmas.begin()+attribINDEX);
            nbSigmas = sigmas.size();
            break;
}
}

//insert at the proper place the (x,y) point in the control point list
void Catmull::insertPoint(double x,double y, int id, double t)
{
    if(id<2 || id>=nbPoint)
        return;

    //insert point P between P1 and P2 (tangent points)
    //P2 = controlPoints[id]
    //PtB = backward tangent at P, PtF = forward tangent at P
    Point P(x,y);
    Point P1, P2, PtB, PtF;
    //compute the tangente and inverse tangent at P(t) from the normal:
    Point Pt = ComputePNormal(t,id-2);
    double temp;
    temp = Pt.getX();
    Pt.setX(Pt.getY());
    Pt.setY(-temp);
    Point Ptinv(-Pt.getX(),-Pt.getY());
    
    //compute the tangent length
    Point ctrlP[4];//resulting bezier : P0 P1modif PtB P for backward tangent, P PtF P2modif P3 for forward tangent
    QVector<Point> data(5);
    QVector<double> uPrime(5);
    Point tHat1,tHat2;
    
    //backward tangent
    //input data point (P0,X1,X2,X3,P), with X1,X2,X3 some curve points between P0 and P
    data[0] = controlPoints[id-2];//P0
    data[1] = ComputeP(t/4.0,id-2);
    data[2] = ComputeP((2.0*t)/4.0,id-2);
    data[3] = ComputeP((3.0*t)/4.0,id-2);
    data[4] = P;
    chord_length_parameterize(&data, &uPrime);
    tHat1 = controlPoints[id-1]-controlPoints[id-2];//P1-P0
    tHat1.normalize();
    tHat2 = Pt;//tangent at P
    estimate_lengths(ctrlP,data,uPrime,tHat1,tHat2);
    //store the modified P1
    P1 = ctrlP[1];
    //store the backward tangent
    PtB = ctrlP[2];
    
    //forward tangent
    //input data point (P,X1,X2,X3,P3), with X1,X2,X3 some curve points between P and P3
    data[0] = P;
    data[1] = ComputeP(t+(1.0-t)/4.0,id-2);
    data[2] = ComputeP(t+(2.0*(1.0-t))/4.0,id-2);
    data[3] = ComputeP(t+(3.0*(1.0-t))/4.0,id-2);
    data[4] = controlPoints[id+1];//P3
    chord_length_parameterize(&data, &uPrime);
    tHat1 = Ptinv;//inverse tangent at P
    tHat2 = controlPoints[id]-controlPoints[id+1];//P2-P3
    tHat2.normalize();
    estimate_lengths(ctrlP,data,uPrime,tHat1,tHat2);
    //store the modified P2
    P2 = ctrlP[2];
    //store the backward tangent
    PtF = ctrlP[1];
    
    //update P1 and P2;
    controlPoints[id-1] = P1;
    controlPoints[id] = P2;
    //insert backward tangent
    controlPoints.insert(controlPoints.begin()+id,PtB);
    //insert the point
    controlPoints.insert(controlPoints.begin()+id,P);
    //insert the forward tangent
    controlPoints.insert(controlPoints.begin()+id,PtF);
    
    //update the number of control points
    nbPoint = controlPoints.size();

    //id+1 to reach the next end point
    //end points: origin P0, destination P3;
    //control points P1, P2
    updateAttribIDs(int (id+1), true);
    updateBoundingBox();
}

//split a curve
//if this is a closed curve, return false and a NULL c2,
//else: return true and init c2 with the remaining curve: c -> c + c2
bool Catmull::split(double x,double y, int id, double t, Catmull* c2)
{
    if(id<2 || id>=nbPoint)
        return false;

    if(!isClosedCurve()){ //open curve
        //insert the splitting point P
        insertPoint(x,y,id,t);
        id++;
        //insert attribute points
        int idLc, idRc, idB;
        int globalID = (id/3)*NB_STEPS;
        idLc = insertAttribPointPos(globalID, COLOR_LEFT);
        idRc = insertAttribPointPos(globalID, COLOR_RIGHT);
        idB = insertAttribPointPos(globalID, BLUR);
        
        //copy the control point following P into the new curve
        c2->controlPoints.clear();
        c2->controlPoints = controlPoints.mid(id);
        c2->nbPoint = c2->controlPoints.size();
        
        //copy the lifetime
        c2->lifetime = lifetime;
        
        //copy attributes into the new curve
        c2->leftColors.clear();
        c2->leftColors = leftColors.mid(idLc);
        c2->leftColorsIDs = leftColorsIDs.mid(idLc);
        c2->nbLeftColors = c2->leftColors.size();
        
        c2->rightColors.clear();
        c2->rightColors = rightColors.mid(idRc);
        c2->rightColorsIDs = rightColorsIDs.mid(idRc);
        c2->nbRightColors = c2->rightColors.size();
        
        c2->sigmas.clear();
        c2->sigmas = sigmas.mid(idB);
        c2->sigmasIDs = sigmasIDs.mid(idB);
        c2->nbSigmas = c2->sigmas.size();
        
        //update attributes ids
        for(int i=0; i<c2->nbLeftColors;i++)
{
            c2->leftColorsIDs[i]-=globalID;
}
        for(int i=0; i<c2->nbRightColors;i++)
{
            c2->rightColorsIDs[i]-=globalID;
}
        for(int i=0; i<c2->nbSigmas;i++)
{
            c2->sigmasIDs[i]-=globalID;
}
        
        //update the new curve's bounding box
        c2->updateBoundingBox();
        
        //remove the duplicated control points from the original curve
        QVector<Point>::iterator lastP;
        lastP = controlPoints.erase(&controlPoints[id+1],&controlPoints[nbPoint-1]);
        controlPoints.erase(lastP);
        nbPoint = controlPoints.size();
        
        //remove attributes
        QVector<QColor>::iterator lastC;
        QVector<int>::iterator lastID;
        lastC = leftColors.erase(&leftColors[idLc+1],leftColors.end());
        lastID = leftColorsIDs.erase(&leftColorsIDs[idLc+1],leftColorsIDs.end());
        nbLeftColors = leftColors.size();
        
        lastC = rightColors.erase(&rightColors[idRc+1],rightColors.end());
        lastID = rightColorsIDs.erase(&rightColorsIDs[idRc+1],rightColorsIDs.end());
        nbRightColors = rightColors.size();
        
        QVector<int>::iterator lastS;
        lastS = sigmas.erase(&sigmas[idB+1],sigmas.end());
        lastID = sigmasIDs.erase(&sigmasIDs[idB+1],sigmasIDs.end());
        nbSigmas = sigmas.size();
        
        updateBoundingBox();
        
        return true;
}
    else{//closed curve
        c2 = NULL;
        
        //open the curve
        setClosedCurve(false);
        
        //insert the splitting point P twice (beginning and end)
        //end
        insertPoint(x,y,id,t);
        id++;
        int globalIDMerge = ((nbPoint-1)/3)*NB_STEPS;
        //insert attribute points
        int idLcEnd, idRcEnd, idBEnd;
        int globalIDEnd = (id/3)*NB_STEPS;
        idLcEnd = insertAttribPointPos(globalIDEnd, COLOR_LEFT);
        idRcEnd = insertAttribPointPos(globalIDEnd, COLOR_RIGHT);
        idBEnd = insertAttribPointPos(globalIDEnd, BLUR);
        //beginning: duplicate the ctrl point and attributes
        controlPoints.insert(id,controlPoints[id]);
        leftColors.insert(idLcEnd,leftColors[idLcEnd]);
        rightColors.insert(idRcEnd,rightColors[idRcEnd]);
        sigmas.insert(idBEnd,sigmas[idBEnd]);
        leftColorsIDs.insert(idLcEnd,leftColorsIDs[idLcEnd]);
        rightColorsIDs.insert(idRcEnd,rightColorsIDs[idRcEnd]);
        sigmasIDs.insert(idBEnd,sigmasIDs[idBEnd]);
        
        //merge the previous beginning/end
        //i.e. : remove the end
        controlPoints.pop_back();
        nbPoint = controlPoints.size();
        leftColors.pop_back();
        leftColorsIDs.pop_back();
        nbLeftColors = leftColors.size();
        rightColors.pop_back();
        rightColorsIDs.pop_back();
        nbRightColors = rightColors.size();
        sigmas.pop_back();
        sigmasIDs.pop_back();
        nbSigmas = sigmas.size();
        
        //update the attribute IDs
        for(int i=0; i<=idLcEnd;i++)
{
            leftColorsIDs[i]+=globalIDMerge-globalIDEnd;
}
        for(int i=0; i<=idRcEnd;i++)
{
            rightColorsIDs[i]+=globalIDMerge-globalIDEnd;
}
        for(int i=0; i<=idBEnd;i++)
{
            sigmasIDs[i]+=globalIDMerge-globalIDEnd;
}
        
        for(int i=idLcEnd+1; i<nbLeftColors;i++)
{
            leftColorsIDs[i]-=globalIDEnd;
}
        for(int i=idRcEnd+1; i<nbRightColors;i++)
{
            rightColorsIDs[i]-=globalIDEnd;
}
        for(int i=idBEnd+1; i<nbSigmas;i++)
{
            sigmasIDs[i]-=globalIDEnd;
}
        
        //set the inserted point as the beginning/end
        //i.e.: translate everything
        //put the end of the list at the beginning
        QVector<Point> ctrlPNew = controlPoints.mid(id+1);
        //add the beginning at the end
        ctrlPNew+=controlPoints.mid(0,id+1);
        //update the control points
        controlPoints = ctrlPNew;
        
        //put the end of the attributes list at the beginning...
        QVector<QColor> leftColorsNew = leftColors.mid(idLcEnd+1);
        leftColorsNew+=leftColors.mid(0,idLcEnd+1);
        leftColors = leftColorsNew;
        
        QVector<QColor> rightColorsNew = rightColors.mid(idRcEnd+1);
        rightColorsNew+=rightColors.mid(0,idRcEnd+1);
        rightColors = rightColorsNew;
        
        QVector<int> sigmasNew = sigmas.mid(idBEnd+1);
        sigmasNew+=sigmas.mid(0,idBEnd+1);
        sigmas = sigmasNew;
        
        //same for the attributes IDs...
        QVector<int> leftColorsIDsNew = leftColorsIDs.mid(idLcEnd+1);
        leftColorsIDsNew+=leftColorsIDs.mid(0,idLcEnd+1);
        leftColorsIDs = leftColorsIDsNew;
        
        QVector<int> rightColorsIDsNew = rightColorsIDs.mid(idRcEnd+1);
        rightColorsIDsNew+=rightColorsIDs.mid(0,idRcEnd+1);
        rightColorsIDs = rightColorsIDsNew;
        
        QVector<int> sigmasIDsNew = sigmasIDs.mid(idBEnd+1);
        sigmasIDsNew+=sigmasIDs.mid(0,idBEnd+1);
        sigmasIDs = sigmasIDsNew;
        
        //move the beginnig away from the end in the direction of tangent
        //compute the tangente and inverse tangent at P(t) from the normal:
        Point Pn = ComputePNormal(0,0);
        Point Pt;
        double temp;
        temp = Pt.getX();
        Pt.setX(Pt.getY());
        Pt.setY(-temp);
        controlPoints[0].setX(controlPoints[0].getX()-0.2*Pt.getX()+0.07*Pn.getX());
        controlPoints[0].setY(controlPoints[0].getY()-0.2*Pt.getY()+0.07*Pn.getY());
        
        return false;
}    
}

//erase the control point at id pos from the control point list
void Catmull::erasePoint(int id)
{
    if(id<0 || id>=nbPoint)
        return;

    controlPoints.erase(controlPoints.begin()+id);
    updateAttribIDs(id, false);

	//special cases: first two point and last two points
	//case 1://the forward tangent of P0; erase P0 with it
	//case 0: //first point
    if(id == 0 || id == 1)
{
		//erase the tangent P1, that has id=0 now
        controlPoints.erase(controlPoints.begin());


		//erase the tangent P2, that has id=0 now
		//because P3 has no backwards tangent anymore;
        controlPoints.erase(controlPoints.begin());


        if(isClosedCurve())
{
            controlPoints[controlPoints.size()-1] = controlPoints[0];
			//copy attributes also
            leftColors[leftColors.size()-1] = leftColors[0];
            rightColors[rightColors.size()-1] = rightColors[0];
            sigmas[sigmas.size()-1] = sigmas[0];

}
}
	//case nbPoint-2://the backward tangent of P(n-1); erase P(n-1) with it
	//case nbPoint-1: //last point
    else if (id == (nbPoint-2) || id == (nbPoint-1))
{
		//erase the tangent P(n-2) [==P2], that has id=last now
        controlPoints.pop_back();

		//erase the tangent P(n-3) [==P1], that has id=last now
		//because P(n-4) has no forward tangent anymore;
        controlPoints.pop_back();

        if(isClosedCurve())
{
            controlPoints[0] = controlPoints[controlPoints.size()-1];
			//copy attributes also
            leftColors[0] = leftColors[leftColors.size()-1];
            rightColors[0] = rightColors[rightColors.size()-1];
            sigmas[0] = sigmas[sigmas.size()-1];
}
}
	//default: middle point, delete without problem
    else{
        switch(id%3)
{
            case 0: //P0 or P3 type
			//delete tangents
			//backward tangent
                controlPoints.erase(controlPoints.begin()+id-1);


			//forward tangent
                controlPoints.erase(controlPoints.begin()+id-1);

                break;
                case 1://P1 type: forward tangent
			//the "brother" tangent
                    controlPoints.erase(controlPoints.begin()+id-2);

			//the control point
                    controlPoints.erase(controlPoints.begin()+id-2);

                    break;
                    case 2://P2 type: backward tangent
			//the control point
                        controlPoints.erase(controlPoints.begin()+id);

			//the "brother" tangent
                        controlPoints.erase(controlPoints.begin()+id);

                        break;
}
}

    nbPoint = controlPoints.size();

    updateBoundingBox();
}

//updateBoundingBox(x, y);
void Catmull::updateBoundingBox()
{
    double x = controlPoints[0].getX();
    double y = controlPoints[0].getY();
    leftUp.setX(x); leftUp.setY(y);
    rightDown.setX(x); rightDown.setY(y);

    for (int i=1; i<nbPoint; i++)
{
        x = controlPoints[i].getX();
        y = controlPoints[i].getY();

        if(x<leftUp.getX()) leftUp.setX(x);
        if(y<leftUp.getY()) leftUp.setY(y);
        if(x>rightDown.getX()) rightDown.setX(x);
        if(y>rightDown.getY()) rightDown.setY(y);
}
}

void Catmull::getBoundingBox(Point& bb_leftUp, Point& bb_rightDown)
{
    bb_leftUp.setX(leftUp.getX()); bb_leftUp.setY(leftUp.getY());
    bb_rightDown.setX(rightDown.getX()); bb_rightDown.setY(rightDown.getY());
}

//set the control point i to the be current moving point
void Catmull::setMovingPoint(int id)
{
    if(id>=0 && id<nbPoint)
        movingPoint = id;
}

//release the current moving point
void Catmull::releaseMovingPoint()
{
    movingPoint=-1;
}

//move the current moving point to (x,y)
//return: true if a point is currently moving 
bool Catmull::movePoint(double x,double y)
{
    if(movingPoint==-1){
        return false;
}

    double x0, y0, xt, yt;
    Point tHat1, tHat2;
    double tHdist, tHnorm;
    double tHmaxDist = 10*PI/180;
	//if movingPoint is 0 or nbPoint-1, special case
    if(movingPoint == 0)
{
        x0 = controlPoints[movingPoint].getX();
        y0 = controlPoints[movingPoint].getY();

        controlPoints[movingPoint].setX(x);
        controlPoints[movingPoint].setY(y);
        x0 = x-x0;
        y0 = y-y0;

        if(!isClosedCurve())
{
            double dist = (controlPoints[0]-controlPoints[nbPoint-1]).L2_norm();
            if(dist<1.0e-2)
                setClosedCurve(true);
}
        if(isClosedCurve())
{
			//move last point
            controlPoints[nbPoint-1].setX(x);
            controlPoints[nbPoint-1].setY(y);
			//move last point's tangent
            xt = controlPoints[nbPoint-2].getX();
            yt = controlPoints[nbPoint-2].getY();
            controlPoints[nbPoint-2].setX(xt+x0);
            controlPoints[nbPoint-2].setY(yt+y0);
}

		//move the tangent
        xt = controlPoints[1].getX();
        yt = controlPoints[movingPoint+1].getY();
        controlPoints[movingPoint+1].setX(xt+x0);
        controlPoints[movingPoint+1].setY(yt+y0);
        updateBoundingBox();
        return true;
}
    /////////////////
    if(movingPoint == nbPoint-1)
{
        x0 = controlPoints[movingPoint].getX();
        y0 = controlPoints[movingPoint].getY();

        controlPoints[movingPoint].setX(x);
        controlPoints[movingPoint].setY(y);
        x0 = x-x0;
        y0 = y-y0;

        if(!isClosedCurve())
{
            double dist = (controlPoints[0]-controlPoints[nbPoint-1]).L2_norm();
            if(dist<1.0e-2)
                setClosedCurve(true);
}
        if(isClosedCurve())
{
			//move last point
            controlPoints[0].setX(x);
            controlPoints[0].setY(y);
			//move last point's tangent
            xt = controlPoints[1].getX();
            yt = controlPoints[1].getY();
            controlPoints[1].setX(xt+x0);
            controlPoints[1].setY(yt+y0);
}
		//move the tangent
        xt = controlPoints[movingPoint-1].getX();
        yt = controlPoints[movingPoint-1].getY();
        controlPoints[movingPoint-1].setX(xt+x0);
        controlPoints[movingPoint-1].setY(yt+y0);
        updateBoundingBox();
        return true;
}

	//if movingPoint is 1 or nbPoint-2, special tangent case
    /////////////////
    if(movingPoint == 1)
{
        x0 = controlPoints[movingPoint].getX();
        y0 = controlPoints[movingPoint].getY();

        controlPoints[movingPoint].setX(x);
        controlPoints[movingPoint].setY(y);

        if(isClosedCurve())
{
            xt = controlPoints[nbPoint-2].getX();
            yt = controlPoints[nbPoint-2].getY();

			//check if tangents have the same direction (if the curve has not an angle at that point
			//for point P(0) == P(n-1) with backward tangent P(n-2) and forward tangent P(1): 
			//P(1)-P(0)
            tHat1.setX(x0-controlPoints[0].getX()); tHat1.setY(y0-controlPoints[0].getY());
			//P(n-1)-P(n-2)
            tHat2.setX(controlPoints[nbPoint-1].getX()-xt); tHat2.setY(controlPoints[nbPoint-1].getY()-yt); //the inverse direction, to avoid a -(x,y)
            tHnorm = tHat2.L2_norm();
            tHat1.normalize(); tHat2.normalize();
            tHdist = (tHat1-tHat2).L2_norm();

            if(tHdist<tHmaxDist)
{
                x0 = controlPoints[0].getX();
                y0 = controlPoints[0].getY();
                tHat2.setX(x0-x); tHat2.setY(y0-y); tHat2.normalize();
                controlPoints[nbPoint-2].setX(x0+tHat2.x*tHnorm);
                controlPoints[nbPoint-2].setY(y0+tHat2.y*tHnorm);
}
}

        updateBoundingBox();
        return true;
}
    /////////////////
    if(movingPoint == nbPoint-2)
{
        x0 = controlPoints[movingPoint].getX();
        y0 = controlPoints[movingPoint].getY();

        controlPoints[movingPoint].setX(x);
        controlPoints[movingPoint].setY(y);

        if(isClosedCurve())
{
            xt = controlPoints[1].getX();
            yt = controlPoints[1].getY();

			//check if tangents have the same direction (if the curve has not an angle at that point
			//for point P(n-1) == P(0) with backward tangent P(n-2) and forward tangent P(1): 
			//P(1)-P(0)
            tHat1.setX(xt-controlPoints[0].getX()); tHat1.setY(yt-controlPoints[0].getY());
			//P(n-1)-P(n-2)
            tHat2.setX(controlPoints[nbPoint-1].getX()-x0); tHat2.setY(controlPoints[nbPoint-1].getY()-y0); //the inverse direction, to avoid a -(x,y)
            tHnorm = tHat1.L2_norm();
            tHat1.normalize(); tHat2.normalize();
            tHdist = (tHat1-tHat2).L2_norm();

            if(tHdist<tHmaxDist)
{
                x0 = controlPoints[nbPoint-1].getX();
                y0 = controlPoints[nbPoint-1].getY();
                tHat1.setX(x0-x); tHat1.setY(y0-y); tHat1.normalize();
                controlPoints[1].setX(x0+tHat1.x*tHnorm);
                controlPoints[1].setY(y0+tHat1.y*tHnorm);
}
}

        updateBoundingBox();
        return true;
}
	//for all others

    switch(movingPoint%3)
{
        case 0: //P0 or P3 type, but not the very first or the very last point
            x0 = controlPoints[movingPoint].getX();
            y0 = controlPoints[movingPoint].getY();

            controlPoints[movingPoint].setX(x);
            controlPoints[movingPoint].setY(y);
            x0 = x-x0;
            y0 = y-y0;
		//move tangents
		//backward tangent
            xt = controlPoints[movingPoint-1].getX();
            yt = controlPoints[movingPoint-1].getY();
            controlPoints[movingPoint-1].setX(xt+x0);
            controlPoints[movingPoint-1].setY(yt+y0);
		//forward tangent
            xt = controlPoints[movingPoint+1].getX();
            yt = controlPoints[movingPoint+1].getY();
            controlPoints[movingPoint+1].setX(xt+x0);
            controlPoints[movingPoint+1].setY(yt+y0);

            updateBoundingBox();
            break;
            case 1: //P1 type; go backward to check for "brother" tangent
                x0 = controlPoints[movingPoint].getX();
                y0 = controlPoints[movingPoint].getY();

                xt = controlPoints[movingPoint-2].getX();
                yt = controlPoints[movingPoint-2].getY();

		//check if tangents have the same direction (if the curve has not an angle at that point
		//for point P3 with backward tangent P2 and forward tangent P4: 
		//P4-P3
                tHat1.setX(x0-controlPoints[movingPoint-1].getX()); tHat1.setY(y0-controlPoints[movingPoint-1].getY());
		//P3-P2
                tHat2.setX(controlPoints[movingPoint-1].getX()-xt); tHat2.setY(controlPoints[movingPoint-1].getY()-yt); //the inverse direction, to avoid a -(x,y)
                tHnorm = tHat2.L2_norm();
                tHat1.normalize(); tHat2.normalize();
                tHdist = (tHat1-tHat2).L2_norm();

                if(tHdist<tHmaxDist)
{
                    x0 = controlPoints[movingPoint-1].getX();
                    y0 = controlPoints[movingPoint-1].getY();
                    tHat2.setX(x0-x); tHat2.setY(y0-y); tHat2.normalize();
                    controlPoints[movingPoint-2].setX(x0+tHat2.x*tHnorm);
                    controlPoints[movingPoint-2].setY(y0+tHat2.y*tHnorm);
}

                controlPoints[movingPoint].setX(x);
                controlPoints[movingPoint].setY(y);
                updateBoundingBox();
                break;
                case 2: //P2 type; go forward to check for "brother" tangent
                    x0 = controlPoints[movingPoint].getX();
                    y0 = controlPoints[movingPoint].getY();

                    xt = controlPoints[movingPoint+2].getX();
                    yt = controlPoints[movingPoint+2].getY();

		//check if tangents have the same direction (if the curve has not an angle at that point
		//for point P3 with backward tangent P2 and forward tangent P4: P4-P3
                    tHat1.setX(xt-controlPoints[movingPoint+1].getX()); tHat1.setY(yt-controlPoints[movingPoint+1].getY());
		//P3-P2
                    tHat2.setX(controlPoints[movingPoint+1].getX()-x0); tHat2.setY(controlPoints[movingPoint+1].getY()-y0); //the inverse direction, to avoid a -(x,y)
                    tHnorm = tHat1.L2_norm();
                    tHat1.normalize(); tHat2.normalize();
                    tHdist = (tHat1-tHat2).L2_norm();

                    if(tHdist<tHmaxDist)
{
                        x0 = controlPoints[movingPoint+1].getX();
                        y0 = controlPoints[movingPoint+1].getY();
                        tHat1.setX(x0-x); tHat1.setY(y0-y); tHat1.normalize();
                        controlPoints[movingPoint+2].setX(x0+tHat1.x*tHnorm);
                        controlPoints[movingPoint+2].setY(y0+tHat1.y*tHnorm);
}

                    controlPoints[movingPoint].setX(x);
                    controlPoints[movingPoint].setY(y);
                    updateBoundingBox();
                    break;
}
    return true;
}

void Catmull::increaseBlur()
{
    if(sigma<MAX_BLUR){
        sigma++;
}
}

void Catmull::decreaseBlur()
{
    if(sigma>1){
        sigma--;
}
}

void Catmull::setBlur(int s)
{
    sigma = s;
}

int Catmull::getBlur()
{
    return sigma;
}

void Catmull::setLifetime(int l)
{
    lifetime = l;
}

QColor Catmull::getLeftColor()
{
    return leftColor;
}

QColor Catmull::getRightColor()
{
    return rightColor;
}

//change attributes globally
//--- blur
void Catmull::setGlobalBlur(int shift_sigma)
{
	nbSigmas = sigmas.size();
    for(int id=0;id<nbSigmas;id++)
{
        int s = sigmas[id]+shift_sigma;
        if(s>MAX_BLUR) s=MAX_BLUR;
        if(s<0) s=0;
        sigmas[id] = s;
}
}

//--- Contrast
void Catmull::increaseContrast()
{
    //compute new color values
    QVector<QColor> leftColorsNew;
    QVector<QColor> rightColorsNew;
    int geomId;
    int idL,idR;
    double tGeom,tR,tL;
    QColor Cl,Cr,ClNew,CrNew;
    double gr,gg,gb;//gradient values
    int r,g,b;
    //left
    for(idL=0;idL<nbLeftColors;idL++)
{
        //compute left and right color
        Cl = leftColors[idL];
        tGeom = ComputeGeomT(leftColorsIDs[idL], geomId);
        tR = ComputeAttribTAttribIndex(tGeom, geomId, idR, COLOR_RIGHT);
        Cr = ComputeCRight(tR, idR);
        //compute the color gradient
        gr = Cr.red()-Cl.red();
        gg = Cr.green()-Cl.green();
        gb = Cr.blue()-Cl.blue();
        r = (int)((double)Cl.red()-0.1*gr+0.5);
        g = (int)((double)Cl.green()-0.1*gg+0.5);
        b = (int)((double)Cl.blue()-0.1*gb+0.5);
        if(r>255) r=255; if(r<0) r=0;
        if(g>255) g=255; if(g<0) g=0;
        if(b>255) b=255; if(b<0) b=0;
        ClNew.setRed(r);
        ClNew.setGreen(g);
        ClNew.setBlue(b);
        leftColorsNew.insert(leftColorsNew.end(),ClNew);
}
    //right
    for(idR=0;idR<nbRightColors;idR++)
{
        //compute left and right color
        Cr = rightColors[idR];
        tGeom = ComputeGeomT(rightColorsIDs[idR], geomId);
        tL = ComputeAttribTAttribIndex(tGeom, geomId, idL, COLOR_LEFT);
        Cl = ComputeCLeft(tL, idL);
        //compute the color gradient
        gr = Cr.red()-Cl.red();
        gg = Cr.green()-Cl.green();
        gb = Cr.blue()-Cl.blue();
        r = (int)((double)Cr.red()+0.1*gr+0.5);
        g = (int)((double)Cr.green()+0.1*gg+0.5);
        b = (int)((double)Cr.blue()+0.1*gb+0.5);
        if(r>255) r=255; if(r<0) r=0;
        if(g>255) g=255; if(g<0) g=0;
        if(b>255) b=255; if(b<0) b=0;
        CrNew.setRed(r);
        CrNew.setGreen(g);
        CrNew.setBlue(b);
        rightColorsNew.insert(rightColorsNew.end(),CrNew);
}

    //update the old color values
    leftColors = leftColorsNew;
    rightColors = rightColorsNew;
}

void Catmull::decreaseContrast()
{
    //compute new color values
    QVector<QColor> leftColorsNew;
    QVector<QColor> rightColorsNew;
    int geomId;
    int idL,idR;
    double tGeom,tR,tL;
    QColor Cl,Cr,ClNew,CrNew;
    double gr,gg,gb;//gradient values
    int r,g,b;
    //left
    for(idL=0;idL<nbLeftColors;idL++)
{
        //compute left and right color
        Cl = leftColors[idL];
        tGeom = ComputeGeomT(leftColorsIDs[idL], geomId);
        tR = ComputeAttribTAttribIndex(tGeom, geomId, idR, COLOR_RIGHT);
        Cr = ComputeCRight(tR, idR);
        //compute the color gradient
        gr = Cr.red()-Cl.red();
        gg = Cr.green()-Cl.green();
        gb = Cr.blue()-Cl.blue();
        r = (int)((double)Cl.red()+0.1*gr+0.5);
        g = (int)((double)Cl.green()+0.1*gg+0.5);
        b = (int)((double)Cl.blue()+0.1*gb+0.5);
        if(r>255) r=255; if(r<0) r=0;
        if(g>255) g=255; if(g<0) g=0;
        if(b>255) b=255; if(b<0) b=0;
        ClNew.setRed(r);
        ClNew.setGreen(g);
        ClNew.setBlue(b);
        leftColorsNew.insert(leftColorsNew.end(),ClNew);
}
    //right
    for(idR=0;idR<nbRightColors;idR++)
{
        //compute left and right color
        Cr = rightColors[idR];
        tGeom = ComputeGeomT(rightColorsIDs[idR], geomId);
        tL = ComputeAttribTAttribIndex(tGeom, geomId, idL, COLOR_LEFT);
        Cl = ComputeCLeft(tL, idL);
        //compute the color gradient
        gr = Cr.red()-Cl.red();
        gg = Cr.green()-Cl.green();
        gb = Cr.blue()-Cl.blue();
        r = (int)((double)Cr.red()-0.1*gr+0.5);
        g = (int)((double)Cr.green()-0.1*gg+0.5);
        b = (int)((double)Cr.blue()-0.1*gb+0.5);
        if(r>255) r=255; if(r<0) r=0;
        if(g>255) g=255; if(g<0) g=0;
        if(b>255) b=255; if(b<0) b=0;
        CrNew.setRed(r);
        CrNew.setGreen(g);
        CrNew.setBlue(b);
        rightColorsNew.insert(rightColorsNew.end(),CrNew);
}

    //update the old color values
    leftColors = leftColorsNew;
    rightColors = rightColorsNew;
}

//switch left and right colors
void Catmull::switchColors()
{
    QVector<QColor> temp;
    QVector<int> tempIDs;

    temp = leftColors;
    tempIDs = leftColorsIDs;
    leftColors = rightColors;
    leftColorsIDs = rightColorsIDs;
    rightColors = temp;
    rightColorsIDs = tempIDs;
    nbLeftColors = leftColors.size();
    nbRightColors = rightColors.size();

}

//--- Smoothing
void Catmull::smooth()
{
    double alpha = 0.3;//smoothing weight: 1 = high smoothing, 0 = no smoothing
    for(int id=1;id<nbPoint-1;id++)
{
        Point P;
        P.setX((1.0-alpha)*controlPoints[id].getX()
                +alpha*(controlPoints[id-1].getX()+controlPoints[id+1].getX())/2.0);
        P.setY((1.0-alpha)*controlPoints[id].getY()
                +alpha*(controlPoints[id-1].getY()+controlPoints[id+1].getY())/2.0);
        controlPoints[id].setX(P.getX());
        controlPoints[id].setY(P.getY());
}
    updateBoundingBox();
}

//get the mean blur of the spline
int Catmull::getMeanBlur()
{
    long meanB=0;
    for(int i=0; i<nbSigmas; i++){
        meanB+=sigmas[i];
}
    meanB/=nbSigmas;
    return (int)(meanB+0.5);
}

//paste a uniform blur on the spline
void Catmull::pasteBlur(int b)
{
    sigmas.clear();
    sigmasIDs.clear();
    //first blur control point
    sigmas.insert(sigmas.end(),b);
    sigmasIDs.insert(sigmasIDs.end(),0);
    //last blur control point
    sigmas.insert(sigmas.end(),b);
    sigmasIDs.insert(sigmasIDs.end(),((nbPoint-1)/3)*NB_STEPS);

    nbSigmas = sigmas.size();
}

void Catmull::setLeftColor(QColor C)
{
    leftColor=C;
}

void Catmull::setRightColor(QColor C)
{
    rightColor=C;
}

int Catmull::getLifetime()
{
    return lifetime;
}

//set the moving point color
void Catmull::setControlPointLeftColor(QColor C, int id)
{
    if(id <0 || id>=nbLeftColors)
        return;

    leftColors[id] = C;
    if(isClosedCurve())
{
        if( id == nbLeftColors-1) 
            leftColors[0] = C;
        if(id == 0) 
            leftColors[nbLeftColors-1] = C;
}
}

//set the moving point color
void Catmull::setControlPointRightColor(QColor C, int id)
{
    if(id <0 || id>=nbRightColors)
        return;

    rightColors[id] = C;
    if(isClosedCurve())
{
        if( id == nbRightColors-1) 
            rightColors[0] = C;
        if(id == 0) 
            rightColors[nbRightColors-1] = C;
}
}

//set the sigma at control point having the id "id" 
void Catmull::setSigma(int sigmaVal, int id)
{
    if(id <0 || id>=nbSigmas)
        return;
    sigmas[id] = sigmaVal;
    if(isClosedCurve())
{
        if( id == nbSigmas-1) 
            sigmas[0] = sigmaVal;
        if(id == 0) 
            sigmas[nbSigmas-1] = sigmaVal;
}
}
//get the sigma at control point having the id "id"
int Catmull::getSigma(int id)
{
    if(id <0 || id>=nbSigmas)
        return -1;
    return sigmas[id];
}

//-------------------------------//
//   Spline Drawing functions    //
//-------------------------------//

void Catmull::drawWithGlobalIDTags()
{
    glLineWidth(lineWidth);
    double t;
    Point P,Pnormal;

    for(int i=0; i<nbPoint-2; i+=3){

		//draw the Catmull curve

		//each piece of the spline is tagged by its first control point; 
		//and by the current step
		//used when adding/erasing end points (P0, P3, P6 ...)
		//used when adding/erasing attribute control points		
        for(t=0.0;t<1;t+=pas){
            glLoadName((i/3+1)*NB_STEPS+t*NB_STEPS); 
            glBegin(GL_LINE_STRIP);
{
                P=ComputeP(t,i);//compute P(t)
                glVertex3f(P.getX(),P.getY(),0.0f);//and draw it
                P=ComputeP(t+pas,i);//compute P(t)
                glVertex3f(P.getX(),P.getY(),0.0f);//and draw it
}
            glEnd();
}
}//end for all point
    glLineWidth(1);
}


//draw the curve
void Catmull::draw(bool tags)
{
    if(lifetime<lifetimeThreshold){
        return;
}

	////draw the polyline attached
	//int size = polLinePoints.size();
	//glColor3f( 1,0,0 );
	//glBegin(GL_LINE_STRIP);
	//for(int i=0;i<size;i++)
	//{
	//	glVertex3f(polLinePoints[i].getX(),polLinePoints[i].getY(),0.0f);
	//}
	//glEnd();
	glColor3fv( splineColor );

    glLineWidth(lineWidth);
    if(nbPoint<4){//if not enough control points: line segment
        glEnable(GL_LINE_STIPPLE);//enable stippling	
        glLineStipple(1,0XF0F0);//stippling mask
        glBegin(GL_LINE_STRIP);
        for(int i=0;i<nbPoint;i++){
            glVertex3f(controlPoints[i].getX(),controlPoints[i].getY(),0.0f);
}
        glEnd();
        glDisable(GL_LINE_STIPPLE);//disable stippling
}
    else{//if enough control point: curve
        double t;
        Point P,Pnormal;

        for(int i=0; i<nbPoint-2; i+=3){
			//each piece of the spline is tagged by its first control point; 
			//used when adding/erasing points
            if(tags) glLoadName(i+3); //P2_id (= i-2)+ 3
			//draw the Catmull curve
            glBegin(GL_LINE_STRIP);
            for(t=0.0;t<=1;t+=pas){
                P=ComputeP(t,i);//compute P(t)
                glVertex3f(P.getX(),P.getY(),0.0f);//and draw it
}
            glEnd();

			//draw the tangent
			//for(t=0.0;t<=1;t+=pas){
			//        P=ComputeP(t,i-4);//compute P(t)
			//        Pnormal=ComputePNormal(t,i-4);//compute P'(t)
			//        glBegin(GL_LINES);
			//            glVertex3f(P.getX(),P.getY(),0.0f);
			//            glVertex3f(P.getX()+0.1*Pnormal.getX(),P.getY()+0.1*Pnormal.getY(),0.0f);
			//        glEnd();
			//}
}
}
    glLineWidth(1);
}

//draw the color and blur control points
void Catmull::drawAttribControlPoints(bool tags)
{
    if(lifetime<lifetimeThreshold){
        return;
}
    if(nbLeftColors<2 || nbRightColors<2 || nbSigmas<2){//if not enough attribute points
        return;
}

    glMatrixMode(GL_MODELVIEW) ;
    glLineWidth(lineWidth);

    const unsigned int tagBlur = 10000;
    const unsigned int tagLeft = 20000;
    const unsigned int tagRight = 30000;

    QColor Cl, Cr;
    double leftR, leftG, leftB;
    double rightR, rightG, rightB;
    Point P, Pnormal;

    GLfloat x, y, radius;
    radius = controlPointsSize/2;

    glActiveTexture(GL_TEXTURE1_ARB);
    glDisable( GL_TEXTURE_RECTANGLE_ARB );
    glActiveTexture(GL_TEXTURE0_ARB);
    glDisable( GL_TEXTURE_RECTANGLE_ARB );


    GLUquadricObj *quadratic;	
    quadratic=gluNewQuadric();			// Create A Pointer To The Quadric Object ( NEW )
    gluQuadricNormals(quadratic, GLU_SMOOTH);	// Create Smooth Normals ( NEW )
    gluQuadricTexture(quadratic, GL_FALSE);		// Create Texture Coords ( NEW )

	//----------------------------------------
    double geomT;
    int geomINDEX = 0;
    
	//left colors
    for(int i=0; i<nbLeftColors; i++){
		//compute colors
		//if closed curve, don't draw color[0]
        if(i>0 || !isClosedCurve())
{
            Cl=leftColors[i];
            leftR=(double)(Cl.red())/255.0;
            leftG=(double)(Cl.green())/255.0;
            leftB=(double)(Cl.blue())/255.0;

            geomT = ComputeGeomT(leftColorsIDs[i], geomINDEX);

            P = ComputeP(geomT, geomINDEX);
            Pnormal=ComputePNormal(geomT, geomINDEX);//compute the normal 
			//draw left point
            if(tags)
                glLoadName(i+tagLeft);
            else //a different tag, needed for right click (insert/erase points)
                glLoadName(leftColorsIDs[i]+NB_STEPS);

            x = P.getX() + 4*radius*Pnormal.getX(); 
            y = P.getY() + 4*radius*Pnormal.getY();
            glPushMatrix();
            glTranslatef(x,y,0);	
            glColor3fv( splineColor );
            gluDisk(quadratic,radius,radius*1.5,32,1);
            glColor3f( leftR, leftG, leftB );
            gluDisk(quadratic,0.0f,radius,32,1);
            glPopMatrix();
}
}

    geomINDEX = 0; geomT = 0;
	//right colors
    for(int i=0; i<nbRightColors; i++){
		//compute colors
        if(i>0 || !isClosedCurve())
{
            Cr=rightColors[i];
            rightR=(double)(Cr.red())/255.0;
            rightG=(double)(Cr.green())/255.0;
            rightB=(double)(Cr.blue())/255.0;

            geomT = ComputeGeomT(rightColorsIDs[i], geomINDEX);

            P = ComputeP(geomT, geomINDEX);
            Pnormal=ComputePNormal(geomT, geomINDEX);//compute the normal 
			//draw left point
            if(tags)
                glLoadName(i+tagRight);
            else //a different tag, needed for right click (insert/erase points)
                glLoadName(rightColorsIDs[i]+NB_STEPS);
            x = P.getX() - 4*radius*Pnormal.getX(); 
            y = P.getY() - 4*radius*Pnormal.getY();
            glPushMatrix();
            glTranslatef(x,y,0);	
            glColor3fv( splineColor );
            gluDisk(quadratic,radius,radius*1.5,32,1);
            glColor3f( rightR, rightG, rightB );
            gluDisk(quadratic,0.0f,radius,32,1);
            glPopMatrix();
}
}

	//blur 
    geomINDEX = 0; geomT = 0;
	//right colors
    for(int i=0; i<nbSigmas; i++){
		//compute blur
        if(i>0 || !isClosedCurve())
{
            float blurC = (float)sigmas[i]/(float)MAX_BLUR;
            geomT = ComputeGeomT(sigmasIDs[i], geomINDEX);

            P = ComputeP(geomT, geomINDEX);
            Pnormal=ComputePNormal(geomT, geomINDEX);//compute the normal 
			//draw left point
            if(tags)
                glLoadName(i+tagBlur);
            else //a different tag, needed for right click (insert/erase points)
                glLoadName(sigmasIDs[i]+NB_STEPS);
            x = P.getX(); 
            y = P.getY();
            glPushMatrix();
            glTranslatef(x,y,0);	
            glColor3fv( splineColor );
            gluDisk(quadratic,radius,radius*1.5,32,1);
            glColor3f( blurC, blurC, blurC );
            gluDisk(quadratic,0.0f,radius,32,1);
            glPopMatrix();
}
}

    glMatrixMode(GL_MODELVIEW) ;
    glColor3f( 0.f, 0.f, 0.f);
    glLineWidth(1);
}


void Catmull::drawControlPoints()
{
    if(lifetime<lifetimeThreshold){
        return;
}

    GLfloat x, y, x_0, y_0, radius;
    glActiveTexture(GL_TEXTURE1_ARB);
    glDisable( GL_TEXTURE_RECTANGLE_ARB );
    glActiveTexture(GL_TEXTURE0_ARB);
    glDisable( GL_TEXTURE_RECTANGLE_ARB );
    glColor3fv( controlPointsColor );
    radius = controlPointsSize/2;

    glMatrixMode(GL_MODELVIEW) ;
    glLineWidth(lineWidth);

    GLUquadricObj *quadratic;	
    quadratic=gluNewQuadric();			// Create A Pointer To The Quadric Object ( NEW )
    gluQuadricNormals(quadratic, GLU_SMOOTH);	// Create Smooth Normals ( NEW )
    gluQuadricTexture(quadratic, GL_FALSE);		// Create Texture Coords ( NEW )

    const unsigned int tagCtrlPoint = 10000;

	//draw the tangents: P1, P2 first, so that P0, P3 will be on top, and the tangent lines below
    int i;
    for(i=1; i<nbPoint; ){
		//P1
        glLoadName(i+tagCtrlPoint);
        x = controlPoints[i].getX(); 
        y = controlPoints[i].getY();

        glColor3fv( tangentPointsColor );
        x_0 = controlPoints[i-1].getX(); 
        y_0 = controlPoints[i-1].getY();
        glPushMatrix();
        glBegin(GL_LINES);
{
            glVertex3f(x_0,y_0,0.0f);
            glVertex3f(x,y,0.0f);
}
        glEnd();
        glTranslatef(x,y,0);
        gluDisk(quadratic,0.0f,radius,32,1);
        glPopMatrix();
		//P2
        i=i+1; //!!!!
        glLoadName(i+tagCtrlPoint);
        x = controlPoints[i].getX(); 
        y = controlPoints[i].getY();		
        if(i<nbPoint-1)
{
            x_0 = controlPoints[i+1].getX(); 
            y_0 = controlPoints[i+1].getY();
            glPushMatrix();
            glBegin(GL_LINES);
{
                glVertex3f(x,y,0.0f);
                glVertex3f(x_0,y_0,0.0f);
}
            glEnd();
}
        glTranslatef(x,y,0);
        gluDisk(quadratic,0.0f,radius,32,1);
        glPopMatrix();

        i+=2;
}
    for(i=0; i<nbPoint; i+=3){
        glLoadName(i+tagCtrlPoint);
        x = controlPoints[i].getX(); 
        y = controlPoints[i].getY();

		//"control point on the curve (P0 or P3)
        glColor3fv( controlPointsColor );
        glLineWidth(3);
        glBegin(GL_LINES);
{
            glVertex3f(x-radius,y-radius,0.0f);
            glVertex3f(x+radius,y+radius,0.0f);
            glVertex3f(x+radius,y-radius,0.0f);
            glVertex3f(x-radius,y+radius,0.0f);
}
        glEnd();
}

    if(nbPoint && ((i-3)!=nbPoint-1) )
        std::cerr<<"!!! i = "<<i<<" and nbPoint = "<<nbPoint<<"\n";
    glMatrixMode(GL_MODELVIEW) ;
    glColor3f( 0.f, 0.f, 0.f);
    glLineWidth(1);
}

//draw bounding box
void Catmull::drawBoundingBox()
{
    if(lifetime<lifetimeThreshold){
        return;
}

    glActiveTexture(GL_TEXTURE1_ARB);
    glDisable( GL_TEXTURE_RECTANGLE_ARB );
    glActiveTexture(GL_TEXTURE0_ARB);
    glDisable( GL_TEXTURE_RECTANGLE_ARB );

    glColor3fv( boundingBoxColor );
    glLineWidth(lineWidth);
    glEnable(GL_LINE_STIPPLE);//enable stippling	
    glLineStipple(1,0XF0F0);//stippling mask
    glBegin(GL_LINE_STRIP);
{
        glVertex3f(leftUp.getX(),leftUp.getY(),0.0f);
        glVertex3f(rightDown.getX(),leftUp.getY(),0.0f);
        glVertex3f(rightDown.getX(),rightDown.getY(),0.0f);
        glVertex3f(leftUp.getX(),rightDown.getY(),0.0f);
        glVertex3f(leftUp.getX(),leftUp.getY(),0.0f);
}
    glEnd();
    glDisable(GL_LINE_STIPPLE);//disable stippling

    glEnable (GL_BLEND); glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glColor4f(0.f,0.f,0.f,0.f);
    glBegin(GL_QUADS);
{
        glVertex3f(leftUp.getX(),leftUp.getY(),0.0f);
        glVertex3f(rightDown.getX(),leftUp.getY(),0.0f);
        glVertex3f(rightDown.getX(),rightDown.getY(),0.0f);
        glVertex3f(leftUp.getX(),rightDown.getY(),0.0f);
}
    glEnd();
    glDisable (GL_BLEND);

    glLineWidth(1);
    glColor3f( 0.f, 0.f, 0.f);
}

//draw the curve with the X normal color
void Catmull::drawNormalX()
{
    if(lifetime<lifetimeThreshold){
        return;
}

    if(nbPoint<4){//if not enough control points
        return;
}
    else{//if enough control point: curve
        glActiveTexture(GL_TEXTURE1_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        glActiveTexture(GL_TEXTURE0_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        double t;
        Point P,Pnormal;
        double gR,gG,gB;
        QColor Cl, Cr;
        int leftColorINDEX = 0;
        int rightColorINDEX = 0;
        double leftColorT;
        double rightColorT;

        glBegin(GL_LINE_STRIP);
        for(int i=0; i<nbPoint-2; i+=3){
			//draw the Catmull curve
            for(t=0.0;t<=1;t+=pas){
                P=ComputeP(t,i);//compute P(t)
                Pnormal=ComputePNormal(t,i);//compute the normal at P(t)

				//compute colors
                leftColorT = ComputeAttribT(t, i, leftColorINDEX, COLOR_LEFT);
                Cl=ComputeCLeft(leftColorT,leftColorINDEX);
                rightColorT = ComputeAttribT(t, i, rightColorINDEX, COLOR_RIGHT);
                Cr=ComputeCRight(rightColorT, rightColorINDEX);

                gR=ComputeGr(Cl.red(), Cr.red());
                gG=ComputeGg(Cl.green(), Cr.green());
                gB=ComputeGb(Cl.blue(), Cr.blue());
                glColor3f(0.5+0.5*gR*Pnormal.getX(),
                          0.5+0.5*gG*Pnormal.getX(),
                          0.5+0.5*gB*Pnormal.getX());//color of the x normal
                glVertex3f(P.getX(),P.getY(),0.0f);//and draw it
}
}
        glEnd();
}
    glColor3f(1.0,1.0,1.0);
}

//draw the curve with the Y normal color
void Catmull::drawNormalY()
{
    if(lifetime<lifetimeThreshold){
        return;
}

    if(nbPoint<4){//if not enough control points
        return;
}
    else{//if enough control point: curve
        glActiveTexture(GL_TEXTURE1_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        glActiveTexture(GL_TEXTURE0_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        double t;
        Point P,Pnormal;
        double gR,gG,gB;
        QColor Cl, Cr;

        int leftColorINDEX = 0;
        int rightColorINDEX = 0;
        double leftColorT;
        double rightColorT;

        glBegin(GL_LINE_STRIP);
        for(int i=0; i<nbPoint-2; i+=3){
			//draw the Catmull curve
            for(t=0.0;t<=1;t+=pas){
                P=ComputeP(t,i);//compute P(t)
                Pnormal=ComputePNormal(t,i);//compute the normal at P(t)

				//compute colors
                leftColorT = ComputeAttribT(t, i, leftColorINDEX, COLOR_LEFT);
                Cl=ComputeCLeft(leftColorT,leftColorINDEX);
                rightColorT = ComputeAttribT(t, i, rightColorINDEX, COLOR_RIGHT);
                Cr=ComputeCRight(rightColorT, rightColorINDEX);

                gR=ComputeGr(Cl.red(), Cr.red());
                gG=ComputeGg(Cl.green(), Cr.green());
                gB=ComputeGb(Cl.blue(), Cr.blue());
                glColor3f(0.5+0.5*gR*Pnormal.getY(),
                          0.5+0.5*gG*Pnormal.getY(),
                          0.5+0.5*gB*Pnormal.getY());//color of the y normal
                glVertex3f(P.getX(),P.getY(),0.0f);//and draw it
}
}
        glEnd();
}
    glColor3f(1.0,1.0,1.0);
}

//deprecated
void Catmull::drawColorConstraints(double zoom)
{
    if(lifetime<lifetimeThreshold){
        return;
}

	//return;
    if(nbPoint<4){//if not enough control points
        return;
}
    else{//if enough control point: curve
        glActiveTexture(GL_TEXTURE1_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        glActiveTexture(GL_TEXTURE0_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        double t;
        Point P1,P2,Pnormal1,Pnormal2;
        QColor Cl1, Cr1, Cl2, Cr2;
        double leftR,leftG,leftB;
        double rightR,rightG,rightB;
        double alpha=3.0;
        double offset=alpha*pixelSize;

		//draw the Catmull curve
        glBegin(GL_LINES);

        int leftColorINDEX1 = 0;
        int rightColorINDEX1 = 0;
        double leftColorT1;
        double rightColorT1;
        int leftColorINDEX2 = 0;
        int rightColorINDEX2 = 0;
        double leftColorT2;
        double rightColorT2;

        for(int i=0; i<nbPoint-2; i+=3){
            P1=ComputeP(0,i);//compute P(0)
            Pnormal1=ComputePNormal(0,i);//compute the normal at P(0)
            //compute colors at t=0
            leftColorT1 = ComputeAttribT(0, i, leftColorINDEX1, COLOR_LEFT);
            Cl1=ComputeCLeft(leftColorT1,leftColorINDEX1);
            rightColorT1 = ComputeAttribT(0, i, rightColorINDEX1, COLOR_RIGHT);
            Cr1=ComputeCRight(rightColorT1, rightColorINDEX1);
            for(t=pas;t<=1;t+=pas){
                P2=ComputeP(t,i);//compute P(t)
                Pnormal2=ComputePNormal(t,i);//compute the normal at P(t)
                
		//compute colors
                leftColorT2 = ComputeAttribT(t, i, leftColorINDEX2, COLOR_LEFT);
                Cl2=ComputeCLeft(leftColorT2,leftColorINDEX2);
                rightColorT2 = ComputeAttribT(t, i, rightColorINDEX2, COLOR_RIGHT);
                Cr2=ComputeCRight(rightColorT2, rightColorINDEX2);

		//right
                rightR=(double)(Cr1.red())/255.0;
                rightG=(double)(Cr1.green())/255.0;
                rightB=(double)(Cr1.blue())/255.0;
                glColor3f(rightR,rightG,rightB);
                glVertex3f(P1.getX()-offset*Pnormal1.getX(),
                           P1.getY()-offset*Pnormal1.getY(),0.0f);
                rightR=(double)(Cr2.red())/255.0;
                rightG=(double)(Cr2.green())/255.0;
                rightB=(double)(Cr2.blue())/255.0;
                glColor3f(rightR,rightG,rightB);
                glVertex3f(P2.getX()-offset*Pnormal2.getX(),
                           P2.getY()-offset*Pnormal2.getY(),0.0f);

		//left
                leftR=(double)(Cl1.red())/255.0;
                leftG=(double)(Cl1.green())/255.0;
                leftB=(double)(Cl1.blue())/255.0;
                glColor3f(leftR,leftG,leftB);
                glVertex3f(P1.getX()+offset*Pnormal1.getX(),
                           P1.getY()+offset*Pnormal1.getY(),0.0f);
                leftR=(double)(Cl2.red())/255.0;
                leftG=(double)(Cl2.green())/255.0;
                leftB=(double)(Cl2.blue())/255.0;
                glColor3f(leftR,leftG,leftB);
                glVertex3f(P2.getX()+offset*Pnormal2.getX(),
                           P2.getY()+offset*Pnormal2.getY(),0.0f);
                
                P1=P2;
                Pnormal1=Pnormal2;
                Cl1=Cl2;
                Cr1=Cr2;
}
}
        glEnd();
}
}

//draw in the stencil buffer to detect overlapping areas
void Catmull::drawOverlap(double zoom)
{
    if(lifetime<lifetimeThreshold){
        return;
}

    if(nbPoint<4){//if not enough control points
        return;
}
    else{//if enough control point: curve
        double t;
        Point P,Pnormal;
        double offset=5.0*pixelSize;

        glBegin(GL_TRIANGLE_STRIP);
        for(int i=0; i<nbPoint-2; i+=3){
			//draw the Catmull curve
            for(t=0;t<=1;t+=pas){
                P=ComputeP(t,i);//compute P(t)
                Pnormal=ComputePNormal(t,i);//compute the normal at P(t)

                glVertex3f(P.getX()-offset*Pnormal.getX(),
                           P.getY()-offset*Pnormal.getY(),0.0f);

                glVertex3f(P.getX()+offset*Pnormal.getX(),
                           P.getY()+offset*Pnormal.getY(),0.0f);
}
}
        glEnd();
}
}

void Catmull::drawColorConstraintsTriangleStrip(double zoom)
{
    if(lifetime<lifetimeThreshold){
        return;
}

	//return;
    if(nbPoint<4){//if not enough control points
        return;
}
    else{//if enough control point: curve
        glActiveTexture(GL_TEXTURE1_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        glActiveTexture(GL_TEXTURE0_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        double t;
        Point P,Pnormal;
        QColor Cl, Cr;
        double leftR,leftG,leftB;
        double rightR,rightG,rightB;
        double alpha=5.0;
        double offset=alpha*pixelSize;

        colorConstraintsProgram.activate();
		//draw the Catmull curve
        glBegin(GL_TRIANGLE_STRIP);

        int leftColorINDEX = 0;
        int rightColorINDEX = 0;
        double leftColorT;
        double rightColorT;

        for(int i=0; i<nbPoint-2; i+=3){
            for(t=0;t<=1;t+=pas){
                P=ComputeP(t,i);//compute P(t)
                Pnormal=ComputePNormal(t,i);//compute the normal at P(t)
                
		//compute colors
                leftColorT = ComputeAttribT(t, i, leftColorINDEX, COLOR_LEFT);
                Cl=ComputeCLeft(leftColorT,leftColorINDEX);
                rightColorT = ComputeAttribT(t, i, rightColorINDEX, COLOR_RIGHT);
                Cr=ComputeCRight(rightColorT, rightColorINDEX);

                rightR=(double)(Cr.red())/255.0;
                rightG=(double)(Cr.green())/255.0;
                rightB=(double)(Cr.blue())/255.0;

                leftR=(double)(Cl.red())/255.0;
                leftG=(double)(Cl.green())/255.0;
                leftB=(double)(Cl.blue())/255.0;

		//right and left colors, stored in texture coordinates 1,2 and 3
                glMultiTexCoord2fARB(GL_TEXTURE1_ARB,rightR,leftR);
                glMultiTexCoord2fARB(GL_TEXTURE2_ARB,rightG,leftG);
                glMultiTexCoord2fARB(GL_TEXTURE3_ARB,rightB,leftB);

		//right
                glMultiTexCoord2fARB(GL_TEXTURE0_ARB,0.0f,1.0/alpha);
                glVertex3f(P.getX()-offset*Pnormal.getX(),
                           P.getY()-offset*Pnormal.getY(),0.0f);

		//left
                glMultiTexCoord2fARB(GL_TEXTURE0_ARB,1.0f,1.0/alpha);
                glVertex3f(P.getX()+offset*Pnormal.getX(),
                           P.getY()+offset*Pnormal.getY(),0.0f);
}
}
        glEnd();
        colorConstraintsProgram.deactivate();
}
}

//draw the amount of blur per pixel
void Catmull::drawBlurAmount()
{
    if(lifetime<lifetimeThreshold){
        return;
}

    if(nbPoint<4){//if not enough control points
        return;
}
    else{//if enough control point: curve
        glActiveTexture(GL_TEXTURE1_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        glActiveTexture(GL_TEXTURE0_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        double t;
        Point P,Pnormal;
        int s;
        int sigmaINDEX = 0;
        double sigmaT;
        double offset=2.0*pixelSize;
        
        glBegin(GL_TRIANGLE_STRIP);
        
        for(int i=0; i<nbPoint-2; i+=3){
	//draw the Catmull curve
            for(t=0.0;t<=1;t+=pas){
                P=ComputeP(t,i);//compute P(t)
                Pnormal=ComputePNormal(t,i);//compute the normal at P(t)
		//compute the blur at P(t)
                sigmaT = ComputeAttribT(t, i, sigmaINDEX, BLUR);
                s = ComputeBlur(sigmaT, sigmaINDEX);
                glColor4f((float)s/(float)MAX_BLUR,(float)s/(float)MAX_BLUR,(float)s/(float)MAX_BLUR,1);//color of the blur
                
                //offset = max(2,s)*pixelSize;

                //right
                glVertex3f(P.getX()-offset*Pnormal.getX(),
                           P.getY()-offset*Pnormal.getY(),0.0f);

		//left
                glVertex3f(P.getX()+offset*Pnormal.getX(),
                           P.getY()+offset*Pnormal.getY(),0.0f);
}
}
        glEnd();
}
    glColor3f(1.0,1.0,1.0);
}

//----------------//
//    Draging    //
//---------------//
void Catmull::setDragPoint(double x, double y)//initiate dragging a selected spline
{
    dragPoint.setX(x);
    dragPoint.setY(y);
}
void Catmull::dragSpline(double x, double y)//drag a selected spline
{
    double dif_x = x - dragPoint.getX();
    double dif_y = y - dragPoint.getY();

	//bounding box drag
    leftUp.setX(leftUp.getX()+dif_x);
    leftUp.setY(leftUp.getY()+dif_y);
    rightDown.setX(rightDown.getX()+dif_x);
    rightDown.setY(rightDown.getY()+dif_y);

    for ( int i=0;i<nbPoint;i++)
{
        controlPoints[i].setX(controlPoints[i].getX()+dif_x);
        controlPoints[i].setY(controlPoints[i].getY()+dif_y);
}

    setDragPoint(x, y);
}

//-----------------//
//    Selection    //
//-----------------//
//set selected param
bool Catmull::isSelected()
{
    return selectedSpline;
}

void Catmull::setSelected(bool selected)
{
    selectedSpline = selected;
}

//----------------//
//     Snake      //
//----------------//
void Catmull::iterSnake()
{
    Point Vs,Vss;//first and second derivative
    Point P, Psave;
    double alpha = 3;//continuity weight (2)
    double beta = 2;//curvature weight (0.8)
    double w_grad = 4;//image weight (4.)
    double Eimg;//image energy
    double E;//total energy
    double Emin;//minimal energy
    int indStart, indEnd;//indices of the beginning and end of the Bezier segment to concider
    int dir_x, dir_y;//direction of the minimal energy
    double x,y;
    double tCtrlp;
    
    QVector<Point> directions(nbPoint);
    //for each control point
    int i;
    if(isClosedCurve())
        i=1;
    else
        i=0;
    for ( int i=1;i<nbPoint;i++)
    {
        indStart = max(0,3*(i/3)-3);
        indEnd = min(nbPoint-1,3*(i/3)+3);
        tCtrlp = (double)(i-3*(i/3))/(double)NB_STEPS;
        
        //init: central direction
	dir_x = 0;
	dir_y = 0;
        Eimg = 0;
        //interior energy:
        Vs = controlPoints[i]-controlPoints[max(0,i-1)];
        Vss = controlPoints[min(nbPoint-1,i+1)]-controlPoints[i]*2.0+controlPoints[max(0,i-1)];
        //image energy:
        for(int k=indStart; k<=indEnd; k+=3){
            //go along the Catmull curve
            for(double t=0.0;t<=1;t+=pas){
                P=ComputeP(t,k);
                x = 0.5*(P.getX()+1);
                y = 1-0.5*(P.getY()+1);
                if(x>0 && x<1 && y>0 && y<1){
                    Eimg += qRed(exampleGradient->pixel((int)(x*exampleGradient->width()),
                                 (int)(y*exampleGradient->height())))/255.0;
                }
                else{
                    Eimg += 1;
                }
            }
        }
        //total energy
        Emin = alpha*Vs.L2_norm()+beta*Vss.L2_norm()+w_grad*Eimg;
        
        //for each possible direction, compute the energy of the snake
        for(int m=-6; m<=6; m++){
            for(int n=-6; n<=6; n++){
		Eimg = 0;
                Psave = controlPoints[i];//save the ctrl point position
                controlPoints[i] = controlPoints[i]+Point((double)m*pixelSize,(double)n*pixelSize);//move the ctrl point
                //interior energy:
                Vs = controlPoints[i]-controlPoints[max(0,i-1)];
                Vss = controlPoints[min(nbPoint-1,i+1)]-controlPoints[i]*2.0+controlPoints[max(0,i-1)];
                //image energy:
		for(int k=indStart; k<=indEnd; k+=3){
                    //go along the Catmull curve
                    for(double t=0.0;t<=1;t+=pas){
                        P=ComputeP(t,k);
                        x = 0.5*(P.getX()+1);
                        y = 1-0.5*(P.getY()+1);
                        if(x>0 && x<1 && y>0 && y<1){
                            Eimg += qRed(exampleGradient->pixel((int)(x*exampleGradient->width()),
                                         (int)(y*exampleGradient->height())))/255.0;
                        }
                        else{
                            Eimg += 1;
                        }
                    }
                }
                //total energy
                E = alpha*Vs.L2_norm()+beta*Vss.L2_norm()+w_grad*Eimg;
                //store the direction that minimize the energy
                if(E<Emin){
                    Emin = E;
                    dir_x = m;
                    dir_y = n;
                }
                //put the ctrl point back to its initial position
                controlPoints[i] = Psave;
            }
        }
        //decision
        directions[i] = Point((double)dir_x*pixelSize,(double)dir_y*pixelSize);
        controlPoints[i] = controlPoints[i]+directions[i];
    }
    /*for ( int i=0;i<nbPoint;i++)
    {
        //update positions
        controlPoints[i] = controlPoints[i]+directions[i];
    }*/
    if(isClosedCurve())
        controlPoints[0]=controlPoints[nbPoint-1];
}


//vectorizes a set of colors; 
QVector<int> Catmull:: vectorizedColorSet(QVector<QColor> colors, QVector<int> colorsIDs, double maxDistance)
{
		int n = (int)colors.size();
	QVector<int> colorsPolLine;
	if(n<=3)
	{
		for(int i=0;i<n;i++)
		{
			colorsPolLine.push_back(i);
		}

		return colorsPolLine;
	}
	//for a line of less than 30 pixels, we only accept 2 colors
	const int max_uni_color = 5;
	const int maxNbBreakPoints = 2+(n/max_uni_color);
	int nbBreakPoints;
	//max acceptable distance between colors, in Lab
	double distMax = maxDistance;
	double  biggestError;
	int biggestErrorPOS;

	biggestError = distMax+1;
	colorsPolLine.push_back(0);
	nbBreakPoints = colorsPolLine.size();
	biggestErrorPOS = n-1;

	while(nbBreakPoints<=maxNbBreakPoints && biggestError>distMax)
	{
		colorsPolLine.push_back(biggestErrorPOS);
		nbBreakPoints ++;
		sort(colorsPolLine.begin(), colorsPolLine.end());
		biggestErrorPOS = computeBiggestError(colorsPolLine, colors, colorsIDs, biggestError);
	}
	std::vector<int> stdvector =  colorsPolLine.toStdVector();
	sort(stdvector.begin(), stdvector.end());
	 QVector<int> vector = QVector<int>::fromStdVector(stdvector);
	return vector;
}
int Catmull::computeBiggestError(QVector<int>& colorPolLine, QVector<QColor> colors, QVector<int> colorsIDs, double& biggestError)
{
		int maxErrorID = colorPolLine[0];
	double maxError = 0;
	int lineSize = colorPolLine.size();
	double x1, x2;
	double t;
	QColor color1, color2, colorInterp;
	LabColor LabColorInterpolated, LabColorValue;

	for(int i=0; i<lineSize-1;i++)
	{
		x1 = colorPolLine[i];
		x2 = colorPolLine[i+1];

		color1 = colors[colorPolLine[i]];
		color2 = colors[colorPolLine[i+1]];

		for(int x=x1; x<x2;x++)
		{
			t = ((double)colorsIDs[x]-(double)colorsIDs[x1])/((double)colorsIDs[x2]-(double)colorsIDs[x1]);
			colorInterp = computeColorLinearInterpolation(t, color1, color2);
			LabColorInterpolated = RGBtoLab(colorInterp);
			LabColorValue = RGBtoLab(colors[x]);
			
			if(labColorDist(LabColorInterpolated, LabColorValue) > maxError)
			{
				maxError = labColorDist(LabColorInterpolated, LabColorValue);
				maxErrorID = x;
			}
		}
	}
	biggestError = maxError;
	return maxErrorID;
}
QColor Catmull::computeColorLinearInterpolation(double t, QColor& color1, QColor& color2)
{
	QColor linInterp;
	linInterp.setRed( color1.red()*(1-t) + color2.red()*t );
	linInterp.setGreen( color1.green()*(1-t) + color2.green()*t );
	linInterp.setBlue( color1.blue()*(1-t) + color2.blue()*t );

	return linInterp;
}
double Catmull::labColorDist(LabColor& color1, LabColor& color2)
{
	double l = pow( (color1.L - color2.L), 2);
	double a = pow( (color1.a - color2.a), 2);
	double b = pow( (color1.b - color2.b), 2);
	return sqrt(l+a+b);
}

LabColor Catmull::RGBtoLab( QColor rgb)
{

	LabColor Lab;


float r, g, b, X, Y, Z, fx, fy, fz, xr, yr, zr;
		float Ls, as, bs;
		float eps = 216.f/24389.f;
		float k = 24389.f/27.f;
		   
		float Xr = 0.964221f;  // reference white D50
		float Yr = 1.0f;
		float Zr = 0.825211f;
		
		// RGB to XYZ
		r = rgb.red()/255.f; //R 0..1
		g = rgb.green()/255.f; //G 0..1
		b = rgb.blue()/255.f; //B 0..1
		
		// assuming sRGB (D65)
		if (r <= 0.04045)
			r = r/12;
		else
			r = (float) pow((r+0.055)/1.055,2.4);
		
		if (g <= 0.04045)
			g = g/12;
		else
			g = (float) pow((g+0.055)/1.055,2.4);
		
		if (b <= 0.04045)
			b = b/12;
		else
			b = (float) pow((b+0.055)/1.055,2.4);
		
		
		X =  0.436052025f*r     + 0.385081593f*g + 0.143087414f *b;
		Y =  0.222491598f*r     + 0.71688606f *g + 0.060621486f *b;
		Z =  0.013929122f*r     + 0.097097002f*g + 0.71418547f  *b;
		
		// XYZ to Lab
		xr = X/Xr;
		yr = Y/Yr;
		zr = Z/Zr;
				
		if ( xr > eps )
			fx =  (float) pow((double)xr, 1/3.);
		else
			fx = (float) ((k * xr + 16.) / 116.);
		 
		if ( yr > eps )
			fy =  (float) pow((double)yr, 1/3.);
		else
		fy = (float) ((k * yr + 16.) / 116.);
		
		if ( zr > eps )
			fz =  (float) pow((double)zr, 1/3.);
		else
			fz = (float) ((k * zr + 16.) / 116);
		
		Ls = ( 116 * fy ) - 16;
		as = 500*(fx-fy);
		bs = 200*(fy-fz);
		
		Lab.L = (int) (2.55*Ls + .5);
		Lab.a = (int) (as + .5); 
		Lab.b = (int) (bs + .5);    

	return Lab;
}

