#include "glslprogram.h"
#include <sstream>
#include <qtextstream.h>
#include <iostream>

#if QT_VERSION >= 0x040000
	#define QTextIStream QTextStream
#endif


using namespace std;

namespace
{
  bool
  sourceAndCompileShader(const char* fileName,GLhandleARB& shader,string& error)
  {
    // Load the source code for the shader
    FILE* file = fopen(fileName,"r");
    if (!file)
    {
      ostringstream os;
      os<<"Cannot open file "<<fileName<<ends;
      error = os.str();
      return false;
    }
    else
    {
      QTextIStream stream(file);
#if QT_VERSION >= 0x040000
		QString s = stream.read(2000000);
		char * sc = strdup(s.toLatin1());
		//cout<<"Shader "<<fileName<<" reads\n"<<sc<<ends;
		GLint lengths = strlen(sc);
      glShaderSourceARB(shader,1,(const char**)&sc,&lengths);
		free(sc);
#else
      QString s = stream.read();
      const char* sc = s.latin1();
      glShaderSourceARB(shader,1,&sc,NULL);
#endif
      // then compile it
      GLint result;
      glCompileShaderARB(shader);
      glGetObjectParameterivARB(shader,GL_COMPILE_STATUS,&result);
      if (result != 1)
      {
	ostringstream os;
	glGetObjectParameterivARB(shader,GL_INFO_LOG_LENGTH,&result);
	char* infoLog = new char[result+1];
	glGetInfoLogARB(shader,result,NULL,infoLog);
	os<<"Shader "<<fileName<<" compilation failure\n"<<infoLog<<ends;
	delete [] infoLog;
	error = os.str();
	return false;
      }
    }
    error="";
    return true;
  }  
}
//************************************************************
// Implementation of GLSLProgram
//************************************************************
GLSLProgram::GLSLProgram()
  : created_(false)
{
}
bool
GLSLProgram::create(const char* vFileName,const char* fFileName)
{
  // First create object to hold shader
  vShader_ = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
  fShader_ = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
  // Create a program,
  program_ = glCreateProgramObjectARB();
  // Attach the shaders to the program
  glAttachObjectARB(program_,vShader_);
  glAttachObjectARB(program_,fShader_);
  created_ = true;
  // Load and link
  return reload(vFileName,fFileName);
}
bool
GLSLProgram::reload(const char* vFileName,const char* fFileName)
{
  if (!created_)
  {
    error_ = "program not created yet";
    return false;
  }
  if (!sourceAndCompileShader(vFileName,vShader_,error_) ||
      !sourceAndCompileShader(fFileName,fShader_,error_))
  {
    return false;
  }
  // Link the program
  glLinkProgramARB(program_);
  GLint result;
  glGetObjectParameterivARB(program_,GL_LINK_STATUS,&result);
  if (result != 1)
  {
    ostringstream os;
    glGetObjectParameterivARB(program_,GL_INFO_LOG_LENGTH,&result);
    char *infoLog = new char[result+1];
    glGetInfoLogARB(program_,result,NULL,infoLog);
    os<<"Program linkage Error\n"
      <<infoLog<<ends;
	 delete [] infoLog;
    error_ = os.str();
    return false;
  }
  error_ = "";
  params_.clear();
  return true;
}
GLint
GLSLProgram::operator[](const string& n) const
{
  map<string,GLint>::const_iterator fter = params_.find(n);
  if (fter != params_.end()) return fter->second;

  GLint p = params_[n] = glGetUniformLocationARB(program_,n.c_str());
  if (p==-1)
  {
    cerr<<"cannot get uniform location of "<<n<<endl;
  }
  else
  {
    cout<<"uniform location of "<<n<<" = "<<p<<endl;
  }
  return p;
}
void
GLSLProgram::activate()
{
  glUseProgramObjectARB(program_);
}
void
GLSLProgram::deactivate()
{
  glUseProgramObjectARB(0);
}
const string&
GLSLProgram::lastError()
{
  return error_;
}
