/****************************************************************************

This file is part of the QGLViewer library.
Copyright (C) 2002, 2003, 2004, 2005, 2006 Gilles Debunne (Gilles.Debunne@imag.fr)
Version 2.2.1-0, released on March 8, 2006.

http://artis.imag.fr/Members/Gilles.Debunne/QGLViewer

libQGLViewer is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

libQGLViewer is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with libQGLViewer; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/

#include "poissonViewer.h"
#include <qimage.h>
#include <QKeyEvent>
#include <QFileDialog>

# include <QMenu>
# include <QMouseEvent>
//#include <qcursor.h>
//#include <qmap.h>

using namespace std;

double Viewer::mouseSelectSize = 0.06;
//-------- Interface slots -----------//

void Viewer::setLeftColor(QColor C)
{
    splines.setLeftColor(C);
    poissonSolver->fastUpdate();
}

void Viewer::setRightColor(QColor C)
{
    splines.setRightColor(C);
    poissonSolver->fastUpdate();
}

void Viewer::setBucketColor(QColor C)
{
    bucketColor = C;
}

void Viewer::setBlur(int sigma)
{
    splines.setBlur(sigma);
    poissonSolver->slowUpdate();
}

//change attributes globally
void Viewer::setGlobalBlur(int sliderBlurVal)
{
    //int shift_sigma = sliderBlurVal - sliderBlurPrevVal;
    //sliderBlurPrevVal = sliderBlurVal;
    splines.setGlobalBlur(sliderBlurVal);
    poissonSolver->slowUpdate();
}

//change a control point color
void Viewer::setControlPointLeftColor(QColor C)
{
    splines.setControlPointLeftColor(C, ATTRIB_ID);
    poissonSolver->fastUpdate();
}
void Viewer::setControlPointRightColor(QColor C)
{
    splines.setControlPointRightColor(C, ATTRIB_ID);
    poissonSolver->fastUpdate();
}

void Viewer::loadXml()
{
    splines.loadXML(QFileDialog::getOpenFileName(this, "Select an xml file", "./", "Vectorized splines (*.xml)"));
    int imageWidth, imageHeight;
    imageWidth = splines.getImageWidth();
    imageHeight = splines.getImageHeight();
    if(imageWidth>0 && imageHeight>0)
    {
        resize(imageWidth, imageHeight);
    }
    zoom=1.0;
    center_x=0.0;
    center_y=0.0;
    poissonSolver->setZoom(zoom);
    poissonSolver->setCenterX(center_x);
    poissonSolver->setCenterY(center_y);
    poissonSolver->slowUpdate();
}

void Viewer::saveXml()
{
    QString name = QFileDialog::getSaveFileName(this, "Choose a filename to save under", "./", "XML (*.xml)");
    if (name.isEmpty()) return;
    splines.saveXML(name);
}

void Viewer::savePng()
{
	Solvenous Solver_solvenous = Solvenous(splines.getImageWidth(),splines.getImageHeight());
   
	QString name = QFileDialog::getSaveFileName(this, "Choose a filename to save under", "./", "Png (*.png)");
    if (name.isEmpty()) return;
    
	Solver_solvenous.setParam(&splines);	
	Solver_solvenous.savePNG(name);
	//poissonSolver->savePng(name);
}

//compute the animation between keyframes
void Viewer::interpolate()
{
    SplineSet s1,s2,sInterpol;

    //select the keyframe files
    QStringList files = QFileDialog::getOpenFileNames(this, "Select xml files", "./", "Vectorized splines (*.xml)");
    if(files.size()<2)
        return;

    QString outputName("output/interpol");

    //set the solver to the default configuration
    poissonSolver->setCenterX(0);
    poissonSolver->setCenterY(0);
    poissonSolver->setZoom(1);

    //interpolate between two keyframes
    QStringList::Iterator it = files.begin();
    s1.loadXML(*it);
    ++it;
    int numFrame=0;
    while(it != files.end()) {
        s2.loadXML(*it);
        for(int t=0;t<=25;t++){
            cout<<"Processing frame "<<numFrame<<"..."<<endl;
            SplineSet::interpolate(&s1,&s2,&sInterpol,(double)t/25.0);
            poissonSolver->setSplines(&sInterpol);
            poissonSolver->slowUpdate();
            QString fileName(outputName);
            fileName.append(QString::number(numFrame));
            fileName.append(".png"); 
            poissonSolver->savePng(fileName);
            numFrame++;
        }
        s1.loadXML(*it);
        ++it;
    }

    //set back the viewer
    poissonSolver->setSplines(&splines);
    poissonSolver->setCenterX(center_x);
    poissonSolver->setCenterY(center_y);
    poissonSolver->setZoom(zoom);
    poissonSolver->slowUpdate();

    cout<<"done!"<<endl;
}

void Viewer::changeManipMode(ManipulationMode manipMode)
{
	//if we quit the select multi mode, deselect all splines
    if(MANIP_MODE == ManipMode_SELECT_MULTI && manipMode!=ManipMode_SELECT_MULTI)
        splines.deselectAll();

    MANIP_MODE = manipMode;
	//if one spline was already selected, add it to the selection vector
    if(MANIP_MODE == ManipMode_SELECT_MULTI)
    {
        int id = splines.getSelected();
        if (id>=0)
        {
            splines.addRemoveToSelection(id);
            splines.setSelected(-1);
        }
    }

    switch (MANIP_MODE){
        case ManipMode_PAN:
        //change the cursor to a cross
        setCursor(Qt::SizeAllCursor);
        break;
        case ManipMode_BUCKET:
        //change the cursor to a finger
        setCursor(Qt::PointingHandCursor);
        break;
        case ManipMode_PICKER:
        //change the cursor to a finger
        setCursor(Qt::WhatsThisCursor);
        break;
        default:
        setCursor(Qt::ArrowCursor);//put back the arrow
        break;
    }
}

void Viewer::zoomIn()
{
    Catmull::zoomIn();
		//the point size for CTRL + LEFT_BUTTON select path
    mouseSelectSize/=2.0;
    zoom*=2.0;
    poissonSolver->setZoom(zoom);
    poissonSolver->slowUpdate();
}

void Viewer::zoomOut()
{
    Catmull::zoomOut();
				//the point size for CTRL + LEFT_BUTTON select path
    mouseSelectSize*=2.0;
    zoom*=0.5;
    poissonSolver->setZoom(zoom);
    poissonSolver->slowUpdate();
}

void Viewer::smoothSplines()
{
    splines.smoothSplines();
    poissonSolver->slowUpdate();
}

void Viewer::stylizeSplines()
{
    splines.stylizeSplines();
    poissonSolver->slowUpdate();
}

void Viewer::increaseContrast()
{
    splines.increaseContrast();
    poissonSolver->slowUpdate();
}

void Viewer::decreaseContrast()
{
    splines.decreaseContrast();
    poissonSolver->slowUpdate();
}

void Viewer::switchColors()
{
    splines.switchColors();
    poissonSolver->slowUpdate();
}

void Viewer::sampleFromExample()
{
	if(exampleIsLoaded){
		splines.sampleFromExample();
		poissonSolver->slowUpdate();
	}
}

void Viewer::iterSnake()
{
	if(exampleIsLoaded){
		splines.iterSnake();
		poissonSolver->fastUpdate();
	}
}

//copy/paste
void Viewer::copySplines()
{
    splines.copySplines();
}

void Viewer::pasteSplines()
{
    double x = (cursor().pos().x())/(0.5*width())-1;
    double y = 1-(cursor().pos().y())/(0.5*height());
    x/=zoom;
    y/=zoom;
    x-= center_x;
    y-= center_y;
    splines.pasteSplines(x,y, center_x, center_y, zoom);
    poissonSolver->slowUpdate();
}

void Viewer::cutSplines()
{
    splines.cutSplines();
    poissonSolver->slowUpdate();
}

void Viewer::copyBlur()
{
    splines.copyBlur();
}

void Viewer::pasteBlur()
{
    splines.pasteBlur();
    poissonSolver->slowUpdate();
}

void Viewer::setLifetimeThreshold(int th)
{
    Catmull::setLifetimeThreshold(th);
    poissonSolver->slowUpdate();
}

// Constructor must call the base class constructor.
Viewer::Viewer()
    : QGLViewer()
{
    internalTimer = new QTimer(this);
    connect(internalTimer, SIGNAL(timeout()), SLOT(updateGL()));
    internalTimer->start(1);

    zoom=1.0;
    center_x = 0.0;
    center_y = 0.0;
    isPanning = false;
    poissonSolver = new Solver(_w, _h);
    bucketColor = QColor(bucketColorInit[0],bucketColorInit[1],bucketColorInit[2]);
    
    display_mode = displayRESULT;
}

Viewer::~Viewer()
{
    delete internalTimer;
    delete poissonSolver;
}

void Viewer::draw()
{

    //background
    if(showExample && exampleIsLoaded){
        //iterate the snake of the selected spline
        //splines.iterSnake();
        //display the example image
        //zoom and pan
        double wz = _w/(2.0*zoom);
        double hz = _h/(2.0*zoom);
        double cx = 0.5*(-center_x+1)*_w;
        double cy = 0.5*(-center_y+1)*_h;
	
        glViewport(0,0,width(), height());
        glMatrixMode(GL_PROJECTION);glLoadIdentity();
        glMatrixMode(GL_MODELVIEW) ;glLoadIdentity();

        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,0);
        glColor3f(1.0,1.0,1.0);
        glActiveTexture(GL_TEXTURE1_ARB);
        glDisable( GL_TEXTURE_RECTANGLE_ARB );
        glActiveTexture(GL_TEXTURE0_ARB);
        glEnable( GL_TEXTURE_RECTANGLE_ARB );
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB,_exampleImgID);
		glBegin(GL_QUADS);
        glTexCoord2f(cx-wz,cy-hz);glVertex2f(-1,-1);
        glTexCoord2f(cx+wz,cy-wz);glVertex2f( 1,-1);
        glTexCoord2f(cx+wz,cy+wz);glVertex2f( 1, 1);
        glTexCoord2f(cx-wz,cy+wz);glVertex2f(-1, 1);
        glEnd();
    }
    else{
		//display the solution
        poissonSolver->draw(width(), height(), display_mode);
    }
    
    if(SHOW_SPLINES)
    {
        glColor3fv(splineColor);
		//zoom and pan
        glMatrixMode(GL_MODELVIEW) ;
        glScalef(zoom,zoom,0);
        glTranslatef(center_x,center_y,0.0);
        splines.draw();
        glMatrixMode(GL_MODELVIEW) ;glLoadIdentity();
        glColor3f(1.0,1.0,1.0);
    }

    if(selectPath.size()>0) //if mouse path is active
    {
		//draw path
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
        glColor4fv(mouseSelectColor);
		//zoom and pan
        glMatrixMode(GL_MODELVIEW) ;
        glScalef(zoom,zoom,0);
        glTranslatef(center_x,center_y,0.0);
		//glPointSize(mouseSelectSize);
        double radius = mouseSelectSize/2;
        double x, y;
        for(int i=0; i<selectPath.size(); i++){

            x = selectPath[i].getX(); 
            y = selectPath[i].getY();
			//glVertex3f(x,y,0.0f);
            glBegin(GL_QUADS);
            {
                glVertex3f(x-radius,y-radius,0.0f);
                glVertex3f(x+radius,y-radius,0.0f);
                glVertex3f(x+radius,y+radius,0.0f);
                glVertex3f(x-radius,y+radius,0.0f);
            }
            glEnd();
        }
        glMatrixMode(GL_MODELVIEW) ;glLoadIdentity();
        glColor3f(1.0,1.0,1.0);
        glDisable(GL_BLEND);
		//glPointSize(1);
    }

    if(drawFreeHandPath.size()>0) //if free hand draw is active
    {
		//draw path
        glColor3fv(splineColor);
			//zoom and pan
        glMatrixMode(GL_MODELVIEW) ;
        glScalef(zoom,zoom,0);
        glTranslatef(center_x,center_y,0.0);
        double x, y;
        glBegin(GL_LINE_STRIP);
        for(int i=0; i<drawFreeHandPath.size(); i++){

            x = drawFreeHandPath[i].getX(); 
            y = drawFreeHandPath[i].getY();
            glVertex3f(x,y,0.0f);
        }
        glEnd();
        glMatrixMode(GL_MODELVIEW) ;glLoadIdentity();
        glColor3f(1.0,1.0,1.0);
		//glPointSize(1);
    }
}

void Viewer::loadExampleImg()
{
    //load the image
    QString file = QFileDialog::getOpenFileName(this, "Select an image file", "./", "Images (*.jpg *.png *.bmp)");
    if (file.isEmpty()) return;
    exampleImg.load(file);
    exampleImg = exampleImg.scaled(_w,_h);
    exampleGrad = exampleImg;
    
    //compute the gradient norm
    for(int i=0; i<exampleGrad.width(); i++){
        exampleGrad.setPixel(i,0,QColor(0,0,0).rgb());
        exampleGrad.setPixel(i,exampleGrad.height()-1,QColor(0,0,0).rgb());
    }
    for(int j=0; j<exampleGrad.height(); j++){
        exampleGrad.setPixel(0,j,QColor(0,0,0).rgb());
        exampleGrad.setPixel(exampleGrad.width()-1,j,QColor(0,0,0).rgb());
    }
    int rx,gx,bx;
    int ry,gy,by;
    int normR, normG, normB;
    int grad;
	float kernel[3][3] = {{-0.25,0,0.25},{-0.5,0,0.5},{-0.25,0,0.25}};
    for(int i=1; i<exampleImg.width()-1; i++){
        for(int j=1; j<exampleImg.height()-1;j++){
			rx=ry=gx=gy=bx=by=0;
			for(int m=-1; m<=1; m++){
				for(int n=-1; n<=1; n++){
					rx += kernel[1+m][1+n]*qRed(exampleImg.pixel(i+m,j+n));
					ry += kernel[1+n][1+m]*qRed(exampleImg.pixel(i+m,j+n));
					gx += kernel[1+m][1+n]*qGreen(exampleImg.pixel(i+m,j+n));
					gy += kernel[1+n][1+m]*qGreen(exampleImg.pixel(i+m,j+n));
					bx += kernel[1+m][1+n]*qBlue(exampleImg.pixel(i+m,j+n));
					by += kernel[1+n][1+m]*qBlue(exampleImg.pixel(i+m,j+n));
				}
			}
            normR = sqrt((float)(rx*rx+ry*ry));
            normG = sqrt((float)(gx*gx+gy*gy));
            normB = sqrt((float)(bx*bx+by*by));
            grad = 255-max(max(normR,normG),normB);
            exampleGrad.setPixel(i,j,qRgb(grad,grad,grad));
        }
    }
	
	//create the image texture
    QImage qglImg = QGLWidget::convertToGLFormat(exampleImg);
    if(exampleIsLoaded){
        glDeleteTextures(1,&_exampleImgID);
    }
    glGenTextures(1,&_exampleImgID);
    glBindTexture( GL_TEXTURE_RECTANGLE_ARB,_exampleImgID);
    glTexImage2D( GL_TEXTURE_RECTANGLE_ARB, 0, 4, qglImg.width(), qglImg.height(), 0, GL_RGBA, GL_UNSIGNED_BYTE, qglImg.bits());
    glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
    glTexParameterf(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
    
    exampleIsLoaded = true;
    showExample = true;
}

void Viewer::init()
{
    glewInit();

    resize(_w,_h);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
	//ortho projection
    glMatrixMode(GL_PROJECTION);glLoadIdentity();
    glMatrixMode(GL_MODELVIEW) ;glLoadIdentity();

    poissonSolver->init();
    poissonSolver->setSplines(&splines);
    poissonSolver->setCenterX(center_x);
    poissonSolver->setCenterY(center_y);
    poissonSolver->setZoom(zoom);
    poissonSolver->slowUpdate();

    splines.init(1.0/(double)_w,_w,_h,poissonSolver->getRefId(),&exampleImg,&exampleGrad);
    exampleIsLoaded = false;
    showExample = false;
    
    //loadExampleImg();

    SHOW_SPLINES = true;
    ATTRIB_ID = 0;
    MANIP_MODE = ManipMode_DISABLED;
}

// add keyboard callback
void Viewer::keyPressEvent(QKeyEvent *e)
{
	//keys that have the same effect for all manipulation modes
	//COPY/PASTE splines
    if( (e->key() == Qt::Key_C ) && ( e->modifiers() == Qt::ControlModifier) ) //CTRL_C
    {
        copySplines();
        return;
    }
    if( (e->key() == Qt::Key_X ) && ( e->modifiers() == Qt::ControlModifier) ) //CTRL_X
    {
        cutSplines();
        return;
    }
    if( (e->key() == Qt::Key_V ) && ( e->modifiers() == Qt::ControlModifier) ) //CTRL_V
    {
        pasteSplines();
        return;
    }
        //copy paste blur
    if( (e->key() == Qt::Key_D ) && ( e->modifiers() == Qt::ControlModifier) ) //CTRL_SHIFT_C
    {
        copyBlur();
        return;
    }
    if( (e->key() == Qt::Key_F ) && ( e->modifiers() == Qt::ControlModifier) ) //CTRL_SHIFT_C
    {
        pasteBlur();
        return;
    }
        //reset
    if( e->key() == Qt::Key_R )
    {
        splines.reset();
        zoom=1.0;
        center_x=0.0;
        center_y=0.0;
        poissonSolver->setZoom(zoom);
        poissonSolver->setCenterX(center_x);
        poissonSolver->setCenterY(center_y);
        poissonSolver->slowUpdate();
        return;
    }
	//debug: multigrid level
    if( e->key() == Qt::Key_P )
    {
        poissonSolver->increaseDeepnessDebug();
        return;
    }
    if( e->key() == Qt::Key_L )
    {
        poissonSolver->decreaseDeepnessDebug();
        return;
    }
    
    //display mode: result (normal display)
    if( e->key() == Qt::Key_F1 )
    {
        display_mode = displayRESULT;
        return;
    }
    if( e->key() == Qt::Key_F2 )
    {
        display_mode = displayCOLOR_CONSTRAINTS;
        return;
    }
    if( e->key() == Qt::Key_F3 )
    {
        display_mode = displayBLUR_CONSTRAINTS;
        return;
    }
    if( e->key() == Qt::Key_F4 )
    {
        display_mode = displayBLUR_MAP;
        return;
    }
    if( e->key() == Qt::Key_F5 )
    {
        display_mode = displayGX;
        return;
    }
    if( e->key() == Qt::Key_F6 )
    {
        display_mode = displayGY;
        return;
    }
    if( e->key() == Qt::Key_F7 )
    {
        display_mode = displayNOTHING;
        return;
    }


	//specific actions, depending on the manipulation mode selected
    switch(MANIP_MODE)
    {
        case ManipMode_SELECT_ONE_GEOMETRY:
        case ManipMode_SELECT_ONE_REDRAW:
        case ManipMode_SELECT_ONE_ATTRIB:
        case ManipMode_BUCKET:
            selectOne_keyPressEvent(e);
            break;

        case ManipMode_SELECT_MULTI:
            selectMulti_keyPressEvent(e);
            break;
        case ManipMode_DRAW_CATMULL:
            drawCatmull_keyPressEvent(e);
            break;
        case ManipMode_DRAW_LINE:
            drawLine_keyPressEvent(e);
            break;
        case ManipMode_PAN:
            pan_keyPressEvent(e);
            break;
        default:
            QGLViewer::keyPressEvent(e);
    }
}

void Viewer::mousePressEvent(QMouseEvent* e)
{   
	//std::cerr<<"(x, y) = ("<<(e->pos()).x()<<", "<<(e->pos()).y()<<")\n";
	//specific actions, depending on the manipulation mode selected
    switch(MANIP_MODE)
    {
        case ManipMode_SELECT_ONE_GEOMETRY:
        case ManipMode_SELECT_ONE_REDRAW:
        case ManipMode_SELECT_ONE_ATTRIB:
        case ManipMode_BUCKET:
            selectOne_mousePressEvent(e);
            break;

        case ManipMode_SELECT_MULTI:
            selectMulti_mousePressEvent(e);
            break;
        case ManipMode_DRAW_CATMULL:
            drawCatmull_mousePressEvent(e);
            break;
        case ManipMode_DRAW_LINE:
            drawLine_mousePressEvent(e);
            break;
        case ManipMode_PAN:
            pan_mousePressEvent(e);;
            break;
        case ManipMode_PICKER:
            pickColor(e);
            break;
        default:
            QGLViewer::mousePressEvent(e);
    }
}

void Viewer::pickColor(QMouseEvent *e)
{
    double x,y;
	x = e->x();
	y = height()-e->y();
		
	double x_d,y_d;
	x_d = x/(0.5*width())-1;
	y_d = (y/height()-1)*2+1;
	
	x_d = 0.5*(x_d+1);
	y_d = 0.5*(y_d+1);
			
    bucketColor = poissonSolver->getPixel(x_d,y_d);
    emit bucketColorChanged(bucketColor);
}

void Viewer::mouseMoveEvent(QMouseEvent *e)
{
	//specific actions, depending on the manipulation mode selected
    switch(MANIP_MODE)
    {
        case ManipMode_SELECT_ONE_GEOMETRY:
        case ManipMode_SELECT_ONE_REDRAW:
        case ManipMode_SELECT_ONE_ATTRIB:
        case ManipMode_BUCKET:
            selectOne_mouseMoveEvent(e);
            break;

        case ManipMode_SELECT_MULTI:
            selectMulti_mouseMoveEvent(e);
            break;
        case ManipMode_DRAW_CATMULL:
            drawCatmull_mouseMoveEvent(e);
            break;
        case ManipMode_DRAW_LINE:
            drawLine_mouseMoveEvent(e);
            break;
        case ManipMode_PAN:
            pan_mouseMoveEvent(e);
            break;
        default:
            QGLViewer::mouseMoveEvent(e);
    }
}

void Viewer::mouseReleaseEvent(QMouseEvent *e)
{
	//specific actions, depending on the manipulation mode selected
    switch(MANIP_MODE)
    {
        case ManipMode_SELECT_ONE_GEOMETRY:
        case ManipMode_SELECT_ONE_REDRAW:
        case ManipMode_SELECT_ONE_ATTRIB:
        case ManipMode_BUCKET:
            selectOne_mouseReleaseEvent(e);
            break;

        case ManipMode_SELECT_MULTI:
            selectMulti_mouseReleaseEvent(e);
            break;
        case ManipMode_DRAW_CATMULL:
            drawCatmull_mouseReleaseEvent(e);
            break;
        case ManipMode_DRAW_LINE:
            drawLine_mouseReleaseEvent(e);
            break;
        case ManipMode_PAN:
            pan_mouseReleaseEvent(e);
            break;
        default:
            QGLViewer::mouseReleaseEvent(e);
    }
}

void Viewer::drawLight (GLenum light, float scale) const
{

}

QString Viewer::helpString() const
{
    QString text("<h2>P o i s s o n  V i e w e r</h2>");
    text += "Spline drawing with color diffusion.";
    return text;
}

QString Viewer::mouseString() const
{
    QString text("<table style=\"width: 100%; text-align: left;\" border=\"1\" cellpadding=\"2\" cellspacing=\"2\">");
    text += "<tbody>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top; \"><center><b>Mouse</b></center>";
    text += "</td>";
    text += "<td style=\"vertical-align: top; \"><center><b>Action</b></center>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Left Button - Drawing mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Draw a spline</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Left Button - Manipulation mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Move a control point, move a tangent point</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Left Button - Color and blur mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Change the color or the blur of an attribute point<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Left Button - Pan mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Pan the viewport<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Shift + Left Button - Multiselection mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Add or remove a spline from the set of selected splines<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Ctrl + Left Button - Multiselection mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Scribble to select splines<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Left Button - Pan mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Pan the viewport<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Right Button - Manipulation mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Insert or delete a control point<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Right Button - Color and blur mode<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Insert or delete an attribute point<br>";
    text += "</td>";
    text += "</tr>";
    text += "</tbody>";
    text += "</table>";
    return text;
}

QString Viewer::keyboardString() const
{
    QString text("<table style=\"width: 100%; text-align: left;\" border=\"1\" cellpadding=\"2\" cellspacing=\"2\">");
    text += "<tbody>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\"><center><b>Key</b></center>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\"><center><b>Action</b></center>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Ctrl+C<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Copy a spline or a set of splines<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Ctrl+V<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Paste a spline or a set of splines</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\vertical-align: top;\">Ctrl+D<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Copy the mean blur of a spline or a set of splines<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">Ctrl+F<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Paste a blur on a spline or a set of splines<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">P<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Debug: display a finer multigrid level<br>";
    text += "</td>";
    text += "</tr>";
    text += "<tr>";
    text += "<td style=\"vertical-align: top;\">L<br>";
    text += "</td>";
    text += "<td style=\"vertical-align: top;\">Debug: display a coarser multigrid level<br>";
    text += "</td>";
    text += "</tr>";
    text += "</tbody>";
    text += "</table>";

    return text;
}
