#ifndef VIEWER_H
#define VIEWER_H

/****************************************************************************

 This file is part of the QGLViewer library.
 Copyright (C) 2002, 2003, 2004, 2005, 2006 Gilles Debunne (Gilles.Debunne@imag.fr)
 Version 2.2.1-0, released on March 8, 2006.

 http://artis.imag.fr/Members/Gilles.Debunne/QGLViewer

 libQGLViewer is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 libQGLViewer is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with libQGLViewer; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*****************************************************************************/
#include "global_param.h"
#include "glslprogram.h"
#include <QGLViewer/qglviewer.h>
#include "spline_set.h"
#include "solver.h"
#include "solvenous.h"

#include <QSlider>

class Viewer : public QGLViewer
{
    Q_OBJECT;

    public slots:
        void setLeftColor(QColor C);
        void setRightColor(QColor C);
        void setBucketColor(QColor C);
        void setBlur(int sigma);
        void setIter(int i){poissonSolver->setIter(i);poissonSolver->fastUpdate();};
        //change attributes globally:
        //blur
        void setGlobalBlur(int shift_sigma);
        //smoothing
        void smoothSplines();
        //stylization
        void stylizeSplines();
		//sample colors in example
		void sampleFromExample();
		//snake
		void iterSnake();
        //contrast
        void increaseContrast();
        void decreaseContrast();
        //lifetime
        void setLifetimeThreshold(int th);
	//change a control point color
        void setControlPointLeftColor(QColor C);
        void setControlPointRightColor(QColor C);
        //switch colors
        void switchColors();
        //copy/paste
        void copySplines();
        void pasteSplines();
        void cutSplines();
        void copyBlur();
        void pasteBlur();

        //show or hide splines; is the splines are not visible, no manipulation can be done
        void showSplines(bool show) {SHOW_SPLINES = show;};
	//show example image
	void showExampleImg(bool show) {showExample = show;};
        //show or hide attribute control points
        void showAttrib(bool show) {splines.setShowAttrib(show);};
        //activate or deactivate the blur
        void activateBlur(bool activate) {poissonSolver->setActivateBlur(activate);poissonSolver->slowUpdate();};
        //activate or deactivate the left or right color
        void activateCLeft(bool cl) {splines.setActivateLeft(cl);};
        void activateCRight(bool cr) {splines.setActivateRight(cr);};
		
        //manipulation mode is changed by the user
        void changeManipMode(ManipulationMode manipMode);

        //load and save
        void loadXml();
        void savePng();
        void saveXml();
        void loadExampleImg();
        //compute the animation between keyframes
        void interpolate();
        //zoom
        void zoomIn();
        void zoomOut();
        
    signals:
        void leftColorChanged(QColor);
        void rightColorChanged(QColor);
        void bucketColorChanged(QColor);
		//demand to change a control point color
        void changeControlPointLeftColor(QColor);
        void changeControlPointRightColor(QColor);
        void blurChanged(int);
    
    public:
        Viewer();
        ~Viewer();
    
    protected :
        virtual void draw();
        virtual void init();
        virtual QString helpString() const;
        virtual QString mouseString() const;
        virtual QString keyboardString() const;
		void drawLight (GLenum light, float scale=1.0f) const;
    
    private :
        virtual void keyPressEvent(QKeyEvent *e);
        virtual void mousePressEvent(QMouseEvent* e);
        virtual void mouseMoveEvent(QMouseEvent *e);
        virtual void mouseReleaseEvent(QMouseEvent *e);

		//manipulation functions
        void selectOne_keyPressEvent(QKeyEvent* e);
        void selectOne_mousePressEvent(QMouseEvent* e);
        void selectOne_leftButton(QMouseEvent* e);
        void selectOne_rightButton(QMouseEvent* e);
        void selectOne_mouseMoveEvent(QMouseEvent* e);
        void selectOne_mouseReleaseEvent(QMouseEvent* e);

        void selectMulti_keyPressEvent(QKeyEvent* e);
        void selectMulti_mousePressEvent(QMouseEvent* e);
        void selectMulti_leftButton(QMouseEvent* e);
        void selectMulti_rightButton(QMouseEvent* e);
        void selectMulti_mouseMoveEvent(QMouseEvent* e);
        void selectMulti_mouseReleaseEvent(QMouseEvent* e);


        void drawCatmull_keyPressEvent(QKeyEvent* e);
        void drawCatmull_mousePressEvent(QMouseEvent* e);
        void drawCatmull_mouseMoveEvent(QMouseEvent* e);
        void drawCatmull_mouseReleaseEvent(QMouseEvent* e);

        void drawLine_keyPressEvent(QKeyEvent* e);
        void drawLine_mousePressEvent(QMouseEvent* e);
        void drawLine_mouseMoveEvent(QMouseEvent* e);
        void drawLine_mouseReleaseEvent(QMouseEvent* e);
    
        void pan_keyPressEvent(QKeyEvent* e);
        void pan_mousePressEvent(QMouseEvent* e);
        void pan_mouseMoveEvent(QMouseEvent* e);
        void pan_mouseReleaseEvent(QMouseEvent* e);
        
        void pickColor(QMouseEvent *e);

		//auxiliary functions for select, defined in manip_selectOne.cpp
        GLint getPickingHits(ManipulationMode splineManip, ClickType clickType, double x, double y, double mouseWindow, GLuint* buff);

        //curves
        SplineSet splines;
    
        //poisson solver
        Solver* poissonSolver;
        
        //paint bucket
        QColor bucketColor;
        
        //example image
        QImage exampleImg;
        QImage exampleGrad;
        bool exampleIsLoaded;
        bool showExample;
        GLuint _exampleImgID;
		
	//blur popup window
        QWidget* blurWidget;
        QSlider* blurSlider;
    
        //image sizes
        const static int _w=GRID_SIZE;
        const static int _h=GRID_SIZE;  

		//flags
        bool SHOW_SPLINES;
        unsigned int ATTRIB_ID;
        ManipulationMode MANIP_MODE;
    
        //timer
        QTimer *internalTimer;

        //zoom and pan
        double zoom;
        double center_x;
        double center_y;
        double pan_x0,pan_x1,pan_y0,pan_y1;
        bool isPanning;

		//select multi path
        QVector<Point> selectPath;
        static double mouseSelectSize; //is the size at which the points on the mouse path are drawn

		//free hand drawing data, in screen coordinates (zoom and pan considered)
        QVector<Point> drawFreeHandPath;
		//zoom and pan not considered ... coordinates with their zoomed value; used for fitting Beziers at the zoomed-in precission
		QVector<Point> drawFreeHandZoomed;
        
        //debug: display mode
        DisplayMode display_mode;
};


#endif // VIEWER_H
