#ifndef SOLVER_H
#define SOLVER_H

#include "spline_set.h"
#include <QVector>
#include <QString>

class Solver
{
    public:
        Solver(int width, int height, int nI=3, double z=1.0, double cx=0.0, double cy=0.0);
        ~Solver();
        void draw(int width, int height, DisplayMode display_mode);
        void fastUpdate();
        void slowUpdate();
        void updateReference();//update the reference buffer with the current solution
        void init();//must be called once OpenGl is all set
        void increaseIter();//increase the number of iterations
        void decreaseIter();//decrease the number of iterations
        void setIter(int i);//set the number of iterations
        void increaseDeepnessDebug();//increase the level used for display (debug)
        void decreaseDeepnessDebug();//decrease the level used for display (debug)
        void setZoom(double z);
        void setCenterX(double cx);
        void setCenterY(double cy);
        void setActivateBlur(bool activate);
        GLuint getRefId();//return the reference buffer id
        void savePng(QString name);//save the current solution as a png file
        void setSplines(SplineSet* s);//param: s: address of the set of splines to draw
        QColor getPixel(double x, double y);//get the pixel color at position (x,y), 0<=x,y<1

    private :
        //curves
        SplineSet* splines;

        //image size
        int _w, _h;
        int* w;
        int* h;

        //offscreen drawing
        void updateMultigrid();
        void drawFirstScaleGradient();//draw and blur the gradient map at first scale
        void initGradients();//draw the gradient pyramid
        void iterJacobiColor();//Poisson - Jacobi iterations on each level of the color pyramid
        void iterJacobiBlur();//Laplace - Jacobi iterations on each level of the blur pyramid
        void reblur();//reblur the color image according to the blur map
        void initColorConstraints(bool border);//draw the color constraints pyramid; border indicate if we have to add constraints on the image border
        void initBlurConstraints();//draw the blur amount constraints pyramid
        void drawBlurConstraints(int d);//draw the blur constraints at scale d
	void drawColorConstraints(int d);//draw the color constraints at scale d

        //shaders
        GLSLProgram restrictProgram;
        GLSLProgram divProgram;
        GLSLProgram jacobiPoissonProgram;
        GLSLProgram jacobiLaplaceProgram;
        GLSLProgram anisotropicDiffusionProgram;
		GLSLProgram restrictColorProgram;
        GLSLProgram blurThresholdProgram;
		
        //buffers
        int deepness;
        int deepness_debug;
        //gradient
        QVector<GLuint> _gxBuffers;
        QVector<GLuint> _gxBufferFBOs;
        QVector<GLuint> _gyBuffers;
        QVector<GLuint> _gyBufferFBOs;
        //divergence
        QVector<GLuint> _divBuffers;
        QVector<GLuint> _divBufferFBOs;
        //solution
        QVector<GLuint> _resultBuffers;
        QVector<GLuint> _resultBufferFBOs;
        //reference image (last stable solution)
        GLuint _referenceBuffer;
        GLuint _referenceBufferFBO;
        //temporary buffers used in jacobi iterations
        QVector<GLuint> _tempJacobiBuffers;
        QVector<GLuint> _tempJacobiBufferFBOs;
        //blur maps
        QVector<GLuint> _blurBuffers;
        QVector<GLuint> _blurBufferFBOs;
        //constraints on the amount of blur per pixel along edges
        QVector<GLuint> _blurAmountBuffers;
        QVector<GLuint> _blurAmountBufferFBOs;
        //color constraints
        QVector<GLuint> _colorBuffers;
        QVector<GLuint> _colorBufferFBOs;
        GLuint _colorStencilBuffer;
        //temporary buffer
        GLuint _tempBuffer;
        GLuint _tempBufferFBO;

        int nbIter;//nbre of iterations

        //zoom and pan
        double zoom;
        double center_x;
        double center_y;

        //activate or deactivate the blur
        bool activateBlur;
	int blurStep;
};

#endif
