#include "spline_set.h"
#include "bezier_fit.h"

#include <QFile>
#include <QDomElement>
#include <QTextStream>
#include <iostream>
#include <float.h>
using namespace std;

SplineSet::SplineSet()
{
   nbSplines = 0;
   splines.reserve(4);
   convex_hull.reserve(4);
   setSelected(-1);
   leftColor = QColor(leftColorInit[0], leftColorInit[1], leftColorInit[2]);
   rightColor = QColor(rightColorInit[0], rightColorInit[1], rightColorInit[2]);
   activateLeft = true;
   activateRight = true;
   sigma=0;
   showAttrib = false;
   draggingSpline = false;

   imageWidth = 512;
   imageHeight = 512;

   leftUp.setX(0); leftUp.setY(0);
   rightDown.setX(0); rightDown.setY(0);
}

SplineSet::~SplineSet()
{

}

//interpolate between s1 and s2: sResult = t*s2 + (1-t)*s1, 0<=t<=1
//precondition: s1 and s2 have the same number of splines and control points
void SplineSet::interpolate(SplineSet* s1, SplineSet* s2, SplineSet* sResult, double t)
{
   sResult->reset();
   sResult->nbSplines = s1->nbSplines;
   sResult->imageWidth = s1->imageWidth;
   sResult->imageHeight = s1->imageHeight;

   sResult->splines.reserve(s1->nbSplines);
   for(int i=0; i<s1->nbSplines; i++){
       Catmull C;
       Catmull::interpolate(&(s1->splines[i]), &(s2->splines[i]), &C, t);
       sResult->splines.insert(sResult->splines.end(),C);
   }
}

void SplineSet::init(double pSize, int wR, int hR, GLuint idR, QImage* img, QImage* grad)
{
   Catmull::initStaticDatas(pSize, wR, hR, idR, img, grad);
}

void SplineSet::reset()
{
   nbSplines=0;
   splines.clear();
   splines.reserve(4);
   setSelected(-1);
   copiedBlur = -1;
}

void SplineSet::insert()
{
   if(nbSplines==0 || splines[nbSplines-1].getNbPoint()>3){
       Catmull C(leftColor,rightColor,sigma);
       splines.insert(splines.end(),C);
       nbSplines = splines.size();
       setSelected(nbSplines-1);
   }
}

void SplineSet::addSpline(Catmull C)
{
   splines.insert(splines.end(),C);
   nbSplines++;
}

void SplineSet::removeSpline(int id)
{
   if(id>=0 && id<nbSplines)
   {
       splines.erase(splines.begin()+id);
       nbSplines--;
   }
}

int SplineSet::getMovingPoint()
{
   if(selected>=0 && selected<nbSplines)
       return (splines[selected].getMovingPoint());
   return -1;
}

//insert the (x,y) point in the control point list of the current spline
void SplineSet::insertPoint(double x,double y, int id, double t)
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].insertPoint(x,y,id,t);
}

void SplineSet::split(double x,double y, int id, double t)
{
   if(selected>=0 && selected<nbSplines){
       Catmull c;
       if(splines[selected].split( x, y, id, t, &c)){
           //if the splitting effectively create a new spline
           addSpline(c);
       }
   }
}

//insert a control point at a position given by globalID in the attribType control points list
// of the current spline
void SplineSet::insertAttribPoint(int globalID, AttribType attribType)
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].insertAttribPoint(globalID,attribType);
}
//erase the control point closest to globalID in the attribType control points list
void SplineSet::eraseAttribPoint(int globalID, AttribType attribType)
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].eraseAttribPoint(globalID,attribType);
}
//erase the control point at id pos from the list of the current spline
void SplineSet::erasePoint(int id)
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].erasePoint(id);
}

//add the (x,y) point in the control point list of the current spline, read the left or right color in the reference buffer if refL or refR is true
void SplineSet::addPoint(double x,double y)
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   splines[selected].addPoint(x,y);
}


//DEBUG
//add the (x,y) point in the polyline point list of the current spline
void SplineSet::addPolLinePoint(double x,double y)
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   splines[selected].addPolLinePoint(x,y);
}

void SplineSet::setClosedCurve(bool _closedCurve)
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   splines[selected].setClosedCurve(_closedCurve);
}
bool SplineSet::isClosedCurve()
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   return splines[selected].isClosedCurve();
}
//******************************************************//
//void SplineSet::addLeftColorPoint(QColor lC)
//{
//    if(selected<0 || selected>=nbSplines){
//        selected = nbSplines-1;
//    }
//    splines[selected].addLeftColorPoint(lC);
//}
//
//void SplineSet::addRightColorPoint(QColor rC)
//{
//    if(selected<0 || selected>=nbSplines){
//        selected = nbSplines-1;
//    }
//    splines[selected].addRightColorPoint(rC);
//}
//******************************************************//
void SplineSet::addLeftColorPoint(int colorID, bool useDefault, QColor lC)//use the default color
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   splines[selected].addLeftColorPoint(colorID, useDefault, lC);
}
void SplineSet::addRightColorPoint(int colorID,  bool useDefault, QColor rC)//use the default color
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   splines[selected].addRightColorPoint(colorID, useDefault, rC);
}
void SplineSet::addBlurPoint(int blurID, bool useDefault, double sigmaBlur)//use the default blur
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   splines[selected].addBlurPoint(blurID, useDefault, sigmaBlur);
}
//************************************************************//
void SplineSet::addLeftColorPoint(bool refL, double x,double y)
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   splines[selected].addLeftColorPoint(refL,x,y);
}
void SplineSet::addRightColorPoint(bool refR, double x,double y)
{
   if(selected<0 || selected>=nbSplines){
       selected = nbSplines-1;
   }
   splines[selected].addRightColorPoint(refR, x, y);
}

//void SplineSet::addBlurPoint(double sigmaBlur)
//{
//    if(selected<0 || selected>=nbSplines){
//        selected = nbSplines-1;
//    }
//    if(sigmaBlur>=0)
//        splines[selected].addBlurPoint(sigmaBlur);
//    else //add the default sigma
//        splines[selected].addBlurPoint(sigma);
//}


//find the closest control point to (x,y) in the selected spline
//and set it as the current moving point
void SplineSet::setMovingPoint(int id)
{
   if(selected>=0 && selected<nbSplines) {
       splines[selected].setMovingPoint(id);
   }
}

//move the current moving point to (x,y) in the selected spline
//return: true if a point is currently moving
bool SplineSet::movePoint(double x,double y)
{
   if(selected>=0 && selected<nbSplines){
       return splines[selected].movePoint(x,y);
   }
   else{
       return false;
   }
}

//release the current moving point
void SplineSet::releaseMovingPoint()
{
   if(selected>=0 && selected<nbSplines){
       splines[selected].releaseMovingPoint();
   }
}

//draw the curves
void SplineSet::draw()
{
   for(int i=0; i<nbSplines; i++)
       splines[i].draw();
   //select one
   if(selected>=0 && selected<nbSplines)
   {
       splines[selected].drawBoundingBox();
       if(showAttrib)
       {
           splines[selected].drawAttribControlPoints();
       }
       else
           splines[selected].drawControlPoints();
   }
   //select multi
   else if(selectedSplines.size()>0)
   {
       drawBoundingBox();
       //drawConvexHull();
       for (int k=0;k<selectedSplines.size();k++)
           splines[selectedSplines[k]].drawControlPoints();
   }
}

//draw the curves with name tages attached
void SplineSet::drawWithNameTags(ManipulationMode splineManip, ClickType clickType)
{

   if(splineManip == ManipMode_SELECT_MULTI)
   {
       switch(clickType)
       {
       case ClickType_LEFT_BUTTON: //drag mode
           glLoadName(0);
           for(int i=0; i<nbSplines; i++){
               if(!splines[i].isSelected())
                   splines[i].draw();
           }
           glLoadName(1);
           drawBoundingBox();
           for (int k=0;k<selectedSplines.size();k++)
               splines[selectedSplines[k]].draw();
           break;
       case ClickType_CTRL_LEFT: //select by mouse moving
       case ClickType_SHIFT_LEFT://select by click
           for(int i=0; i<nbSplines; i++){
               glLoadName(i);
               splines[i].draw();
           }
           break;
       default:
           ;
       }
       return;
   }
   //if no selection
   if(((splineManip == ManipMode_SELECT_ONE_GEOMETRY ||
       splineManip == ManipMode_SELECT_ONE_ATTRIB ||
       splineManip == ManipMode_BUCKET ||
       splineManip == ManipMode_SELECT_ONE_REDRAW)&& (selected <0 || selected>=nbSplines)))
   {
       for(int i=0; i<nbSplines; i++){
           glLoadName(i);
           splines[i].draw();

       }
       return;
   }

   //if a curve is selected, draw all curves but that one under a same name
   if(clickType == ClickType_LEFT_BUTTON)
   {
       if(splineManip == ManipMode_SELECT_ONE_GEOMETRY)
       {
           glLoadName(0);
           for(int i=0; i<nbSplines; i++){
               if(selected != i)
                   splines[i].draw();
           }
           glLoadName(1);
           splines[selected].drawBoundingBox();
           glLoadName(2);
           splines[selected].draw();
           splines[selected].drawControlPoints();

           return;
       }

       if(splineManip == ManipMode_SELECT_ONE_ATTRIB || splineManip == ManipMode_BUCKET)
       {
           glLoadName(0);
           for(int i=0; i<nbSplines; i++){
               if(selected != i)
                   splines[i].draw();
           }
           glLoadName(1);
           splines[selected].drawBoundingBox();
           glLoadName(2);
           splines[selected].draw();
           splines[selected].drawAttribControlPoints(true);
       }
       return;
   }
   //if(clickType == ClickType_RIGHT_BUTTON)
   if(splineManip == ManipMode_SELECT_ONE_GEOMETRY)
   {
       glLoadName(0);
       for(int i=0; i<nbSplines; i++){
           if(selected != i)
               splines[i].draw();
       }
       glLoadName(1);
       splines[selected].drawBoundingBox();
       //draw the curve using the global IDs as tags;
       splines[selected].drawWithGlobalIDTags();
       //draw the control point with control point IDs as tags
       splines[selected].drawControlPoints();
       return;
   }
   if(splineManip == ManipMode_SELECT_ONE_ATTRIB || splineManip == ManipMode_BUCKET)
   {
       glLoadName(0);
       for(int i=0; i<nbSplines; i++){
           if(selected != i)
               splines[i].draw();
       }
       glLoadName(1);
       splines[selected].drawBoundingBox();
       //draw the curve using the global IDs as tags;
       splines[selected].drawWithGlobalIDTags();
       //draw AttribControlPoints with globalIDs as tags
       splines[selected].drawAttribControlPoints();
       return;
   }
}


//draw the curves with the X derivative color
void SplineSet::drawNormalX()
{
   for(int i=0; i<nbSplines; i++){
       splines[i].drawNormalX();
   }
}

//draw the curves with the Y derivative color
void SplineSet::drawNormalY()
{
   for(int i=0; i<nbSplines; i++){
       splines[i].drawNormalY();
   }
}

//draw the amount of blur per pixel
void SplineSet::drawBlurAmount()
{
   glClearColor(0.0f,0.0f,0.0f,0.0f);
   glClear(GL_COLOR_BUFFER_BIT);

   for(int i=0; i<nbSplines; i++){
       splines[i].drawBlurAmount();
   }
}

//draw the curves with the color constraints on each side
void SplineSet::drawColorConstraints(double zoom)
{
   //draw the overlapping areas in the stencil buffer
   drawOverlap(zoom);

   glClearColor(0.0f,0.0f,0.0f,0.0f);
   glClear(GL_COLOR_BUFFER_BIT);

   //stencil test to not draw color constraints on overlapping areas
   glEnable(GL_STENCIL_TEST);
   glStencilFunc(GL_GREATER,2,~0);
   glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);

   glActiveTexture(GL_TEXTURE1_ARB);
   glDisable( GL_TEXTURE_RECTANGLE_ARB );
   glActiveTexture(GL_TEXTURE0_ARB);
   glDisable( GL_TEXTURE_RECTANGLE_ARB );

   for(int i=0; i<nbSplines; i++){
       splines[i].drawColorConstraintsTriangleStrip(zoom);
   }

   glDisable(GL_STENCIL_TEST);
}

//draw in the stencil buffer to detect overlapping areas
void SplineSet::drawOverlap(double zoom)
{
   //set the stencil
   glEnable(GL_STENCIL_TEST);
   glClearStencil(0);
   glClear(GL_STENCIL_BUFFER_BIT);
   glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
   glStencilFunc(GL_ALWAYS,1,~0);
   //disable the color drawing
   glColorMask(0,0,0,0);

   for(int i=0; i<nbSplines; i++){
       splines[i].drawOverlap(zoom);
   }
   glDisable(GL_STENCIL_TEST);

   //enable the color drawing
   glColorMask(1,1,1,1);
}

//draw the spline control points of the selected curve
void SplineSet::drawControlPoints()
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].drawControlPoints();
}

//increase the blur of the selected spline
void SplineSet::increaseBlur()
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].increaseBlur();
}

//decrease the blur of the selected spline
void SplineSet::decreaseBlur()
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].decreaseBlur();
}

//set the blur of the selected spline
void SplineSet::setBlur(int s)
{
   sigma = s;
   if(selected>=0 && selected<nbSplines){
       splines[selected].setBlur(sigma);
   }
}

//change the left color of the selected spline
void SplineSet::setLeftColor(QColor C)
{
   if(selected>=0 && selected<nbSplines){
       splines[selected].setLeftColor(C);
   }
   leftColor=C;
}

//change the right color of the selected spline
void SplineSet::setRightColor(QColor C)
{
   if(selected>=0 && selected<nbSplines){
       splines[selected].setRightColor(C);
   }
   rightColor=C;
}

//change attributes globally
//blur
void SplineSet::setGlobalBlur(int shift_sigma)
{
   //multiple selection
   for(int k=0;k<selectedSplines.size();k++)
   {
       splines[selectedSplines[k]].setGlobalBlur(shift_sigma);
   }
   //single selection
   if(selected>=0 && selected<nbSplines){
       splines[selected].setGlobalBlur(shift_sigma);
   }
}

void SplineSet::smoothSplines()
{
   for(int k=0;k<selectedSplines.size();k++)
   {
       splines[selectedSplines[k]].smooth();
   }
   //single selection
   if(selected>=0 && selected<nbSplines){
       splines[selected].smooth();
   }

   updateBoundingBox();
   updateConvexHull();
}

//activate or deactivate the stylization on the selected splines
void SplineSet::stylizeSplines()
{
   for(int k=0;k<selectedSplines.size();k++)
   {
       splines[selectedSplines[k]].stylize();
   }
   //single selection
   if(selected>=0 && selected<nbSplines){
       splines[selected].stylize();
   }
}

//contrast
void SplineSet::increaseContrast()
{
   for(int k=0;k<selectedSplines.size();k++)
   {
       splines[selectedSplines[k]].increaseContrast();
   }
   //single selection
   if(selected>=0 && selected<nbSplines){
       splines[selected].increaseContrast();
   }
}

void SplineSet::decreaseContrast()
{
   for(int k=0;k<selectedSplines.size();k++)
   {
       splines[selectedSplines[k]].decreaseContrast();
   }
   //single selection
   if(selected>=0 && selected<nbSplines){
       splines[selected].decreaseContrast();
   }
}

//switch left and right colors of the selected splines
void SplineSet::switchColors()
{
   for(int k=0;k<selectedSplines.size();k++)
   {
       splines[selectedSplines[k]].switchColors();
   }
   //single selection
   if(selected>=0 && selected<nbSplines){
       splines[selected].switchColors();
   }
}

//change the left color of the moving point for the selected spline
void SplineSet::setControlPointLeftColor(QColor C, int id)
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].setControlPointLeftColor(C, id);
}

//change the right color of the moving point for the selected spline
void SplineSet::setControlPointRightColor(QColor C, int id)
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].setControlPointRightColor(C, id);
}

//set the sigma at control point having the id "id" in the selected spline
void SplineSet::setSigma(int sigma, int id)
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].setSigma(sigma, id);
}
//get the sigma at control point having the id "id" in the selected spline
int SplineSet::getSigma(int id)
{
   if(selected>=0 && selected<nbSplines)
       return splines[selected].getSigma(id);
   else
       return -1;
}

//get the blur of the selected spline
int SplineSet::getBlur()
{
   return sigma;
}

//get the left color of the selected spline
QColor SplineSet::getLeftColor()
{
   return leftColor;
}

//get the right color of the selected spline
QColor SplineSet::getRightColor()
{
   return rightColor;
}

//get the left color of  moving point in the selected spline
QColor SplineSet::getControlPointLeftColor(int id)
{
   if(selected>=0 && selected<nbSplines)
       return splines[selected].getLeftColor(id);
   return leftColor;
}
//get the right color of  moving point in the selected spline
QColor SplineSet::getControlPointRightColor(int id)
{
   if(selected>=0 && selected<nbSplines)
       return splines[selected].getRightColor(id);
   return rightColor;
}

/**************************************************************************************************/
// SELECTION
void SplineSet::setSelected(int id)
{
   if(id>=0 && id<nbSplines)
   {
       selected = id;
       leftColor = splines[selected].getLeftColor();
       rightColor = splines[selected].getRightColor();
       sigma = splines[selected].getBlur();
   }
   else
       selected = -1;
}

int SplineSet::getSelected()
{
   return selected;
}

//select multi
void SplineSet::addRemoveToSelection(int id){
   switch(splines[id].isSelected())
   {
   case false://add
       addToSelection(id);
       break;
   case true:
       removeFromSelection(id);
       break;
   }
}

void SplineSet::addToSelection(int id){
   if(!splines[id].isSelected())
   {
       splines[id].setSelected(true);
       selectedSplines.push_back(id);
   }

   updateBoundingBox();
   updateConvexHull();
}
void SplineSet::removeFromSelection(int id){
   if(splines[id].isSelected())
   {
       splines[id].setSelected(false);
       for(int k=0;k<selectedSplines.size();k++)
       {
           if(selectedSplines[k] == id)
           {
               selectedSplines.erase(selectedSplines.begin()+k);
               break;
           }
       }
   }
   updateBoundingBox();
   updateConvexHull();
}

void SplineSet::deselectAll()
{
   for(int k=0;k<nbSplines;k++)
   {
       splines[k].setSelected(false);
   }
   selectedSplines.clear();
   if(selected>=0 && selected<nbSplines)
       setSelected(-1);
}

//Compute the convex hull of the set of attribute points
float isLeft(Point P1, Point P0, Point P2 ) {
   return (P1.getX() - P0.getX())*(P2.getY() - P0.getY()) - (P2.getX() - P0.getX())*(P1.getY() - P0.getY());
}

void SplineSet::updateConvexHull()
{
   //build the list of control points of the selected splines
   QVector<Point> V(0);
   for(int i=0;i<selectedSplines.size();i++){
       V+=splines[selectedSplines[i]].getControlPoints();
   }
   // initialize a vector D[] from bottom to top so that the
   // 1st three vertices of V[] are a counterclockwise triangle
   int n = (int)V.size();

   QVector<Point> D(2*n+1);
   int bot = n-2, top = bot+3;   // initial bottom and top deque indices
   D[bot] = D[top] = V[2];       // 3rd vertex is at both bot and top
   if (isLeft(V[0], V[1], V[2]) > 0) {
       D[bot+1] = V[0];
       D[bot+2] = V[1];          // ccw vertices are: 2,0,1,2
   } else {
       D[bot+1] = V[1];
       D[bot+2] = V[0];          // ccw vertices are: 2,1,0,2
   }

   // compute the hull on the vector D[]
   for (int i=3; i < n; i++) {   // process the rest of vertices
       // test if next vertex is inside the deque hull
       if ((isLeft(D[bot], D[bot+1], V[i]) > 0) &&
           (isLeft(D[top-1], D[top], V[i]) > 0) ) {
               continue;         // skip an interior vertex
       }
       // incrementally add an exterior vertex to the deque hull
       // get the rightmost tangent at the deque bot
       while (isLeft(D[bot], D[bot+1], V[i]) <= 0) {
           ++bot;                // remove bot of deque
       }
       D[--bot] = V[i];          // insert V[i] at bot of deque
       // get the leftmost tangent at the deque top
       while (isLeft(D[top-1], D[top], V[i]) <= 0) {
           --top;                // pop top of deque
       }
       D[++top] = V[i];          // push V[i] onto top of deque
   }


   // copy deque D[] to the output hull array and compute arc-length
   convex_hull.clear();
   convex_hull.push_back(D[bot]);
   for (int h=1; h <= (top-bot); h++){
       convex_hull.push_back(D[bot+h]);
   }
}

//multiple selection bounding box
void SplineSet::updateBoundingBox(){
   //select One/select multi is a big mix-up thing. Should be cleaned up :)
   if(selected>=0 && selected<nbSplines)
   {
       splines[selected].getBoundingBox(leftUp, rightDown);
       return;
   }

   //select multi
   if(!selectedSplines.size())
       return;

   Point bb_leftUp, bb_rightDown;
   double x, y;
   splines[selectedSplines[0]].getBoundingBox(leftUp, rightDown);
   for(int k=1;k<selectedSplines.size();k++)
   {
       splines[selectedSplines[k]].getBoundingBox(bb_leftUp, bb_rightDown);
       x = bb_leftUp.getX();
       y = bb_leftUp.getY();

       if(x<leftUp.getX()) leftUp.setX(x);
       if(y<leftUp.getY()) leftUp.setY(y);
       if(x>rightDown.getX()) rightDown.setX(x);
       if(y>rightDown.getY()) rightDown.setY(y);

       x = bb_rightDown.getX();
       y = bb_rightDown.getY();

       if(x<leftUp.getX()) leftUp.setX(x);
       if(y<leftUp.getY()) leftUp.setY(y);
       if(x>rightDown.getX()) rightDown.setX(x);
       if(y>rightDown.getY()) rightDown.setY(y);
   }
}

//repeat of Catmull::drawBoundingBox;
//should just use one...when code is cleaned
void SplineSet::drawBoundingBox(){
   glActiveTexture(GL_TEXTURE1_ARB);
   glDisable( GL_TEXTURE_RECTANGLE_ARB );
   glActiveTexture(GL_TEXTURE0_ARB);
   glDisable( GL_TEXTURE_RECTANGLE_ARB );

   glColor3fv( boundingBoxColor );
   glLineWidth(1);
   glEnable(GL_LINE_STIPPLE);//enable stippling       glLineStipple(1,0XF0F0);//stippling mask
   glBegin(GL_LINE_STRIP);
   {
       glVertex3f(leftUp.getX(),leftUp.getY(),0.0f);
       glVertex3f(rightDown.getX(),leftUp.getY(),0.0f);
       glVertex3f(rightDown.getX(),rightDown.getY(),0.0f);
       glVertex3f(leftUp.getX(),rightDown.getY(),0.0f);
       glVertex3f(leftUp.getX(),leftUp.getY(),0.0f);
   }
   glEnd();
   glDisable(GL_LINE_STIPPLE);//disable stippling

   glEnable (GL_BLEND); glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glColor4f(0.f,0.f,0.f,0.f);
   glBegin(GL_QUADS);
   {
       glVertex3f(leftUp.getX(),leftUp.getY(),0.0f);
       glVertex3f(rightDown.getX(),leftUp.getY(),0.0f);
       glVertex3f(rightDown.getX(),rightDown.getY(),0.0f);
       glVertex3f(leftUp.getX(),rightDown.getY(),0.0f);
   }
   glEnd();
   glDisable (GL_BLEND);

   glLineWidth(1);
   glColor3f( 0.f, 0.f, 0.f);
}

void SplineSet::drawConvexHull(){
   glActiveTexture(GL_TEXTURE1_ARB);
   glDisable( GL_TEXTURE_RECTANGLE_ARB );
   glActiveTexture(GL_TEXTURE0_ARB);
   glDisable( GL_TEXTURE_RECTANGLE_ARB );

   glColor3fv( boundingBoxColor );
   glLineWidth(1);
   glEnable(GL_LINE_STIPPLE);//enable stippling       glLineStipple(1,0XF0F0);//stippling mask
   glBegin(GL_LINE_LOOP);
   for(int i=0; i<convex_hull.size(); i++){
       glVertex3f(convex_hull[i].getX(),convex_hull[i].getY(),0.0f);
   }
   glEnd();
   glDisable(GL_LINE_STIPPLE);//disable stippling

   glLineWidth(1);
   glColor3f( 0.f, 0.f, 0.f);
}

//delete spline in SELECT_MULTI mode
void SplineSet::deleteSelection(){
   /*if(selected>=0 && selected<nbSplines)
   {
   splines.erase(splines.begin()+selected);
   nbSplines--;
   setSelected(-1);
   return;
   }*/
   QVector<Catmull>::iterator splinesId;
   for(splinesId = splines.begin(); splinesId!=splines.end(); splinesId++)
   {
       if(splinesId->isSelected())
       {
           splines.erase(splinesId);
           splinesId--;
       }
   }
   nbSplines = splines.size();
   selectedSplines.clear();
}

/********** Cpoy and Paste *******************************/
void SplineSet::copySplines()
{
   //select one
   if(selected>=0 && selected<nbSplines)
   {
       copiedSplines.clear();
       copiedSplines.push_back(splines[selected]);
       copiedSplines.last().setSelected(false);
   }

   //select multi
   else if(!selectedSplines.isEmpty())
   {
       copiedSplines.clear();
       for(int k=0;k<selectedSplines.size();k++)
       {
           copiedSplines.push_back(splines[selectedSplines[k]]);
           copiedSplines.last().setSelected(false);
       }
   }
}
void SplineSet::cutSplines()
{
   copySplines();
   //select one
   if(selected>=0 && selected<nbSplines)
   {
       removeSpline(selected);
       setSelected(-1);
   }
   //select multi
   else if(!selectedSplines.isEmpty())
   {
       deleteSelection();
   }
}
//I'll have to modify this to accept "outside" paste
void SplineSet::pasteSplines(double x, double y, double center_x, double center_y, double zoom)
{
   Point bb_leftUp, bb_rightDown;
   double x_c, y_c;
   double x_d, y_d;
   int k;

   switch(copiedSplines.size())
   {
   case 0: //no spline to insert
       break;
   case 1://only one spline, assume mode is SELECT_ONE
       addSpline(copiedSplines[0]);
       setSelected(nbSplines-1);
       splines[selected].getBoundingBox(bb_leftUp, bb_rightDown);
       x_c = bb_leftUp.getX()+(bb_rightDown.getX() - bb_leftUp.getX())/2;
       y_c = bb_leftUp.getY()+(bb_rightDown.getY() - bb_leftUp.getY())/2;

       x_d = x_c+0.3*x_c;
       y_d = y_c+0.3*y_c;

       x_c/=zoom;        x_d/=zoom;
       y_c/=zoom;        y_d/=zoom;
       x_c-= center_x;   x_d-= center_x;
       y_c-= center_y;   y_d-= center_y;

       setDragPoint(x_c, y_c, true);

       dragSpline(x_d, y_d);
       setDragPoint(0,0,false);
       break;
   default://multiple splines
       deselectAll();
       for(k=0;k<copiedSplines.size();k++)
       {
           addSpline(copiedSplines[k]);
           addToSelection(nbSplines-1); //updatesBoundingBox
       }

       x_c = leftUp.getX()+(rightDown.getX() - leftUp.getX())/2;
       y_c = leftUp.getY()+(rightDown.getY() - leftUp.getY())/2;

       x_d = x_c+0.3*x_c;
       y_d = y_c+0.3*y_c;

       x_c/=zoom;        x_d/=zoom;
       y_c/=zoom;        y_d/=zoom;
       x_c-= center_x;   x_d-= center_x;
       y_c-= center_y;   y_d-= center_y;

       setDragPoint(x_c, y_c, true);
       dragSpline(x_d, y_d);
       setDragPoint(0,0,false);
   }
}

//copy the mean blur of the selected splines
void SplineSet::copyBlur()
{
   copiedBlur = -1;
   if(!selectedSplines.isEmpty()){//multiple selection
       copiedBlur = 0;
       for(int k=0;k<selectedSplines.size();k++)
       {
           copiedBlur += splines[selectedSplines[k]].getMeanBlur();
       }
       copiedBlur/=selectedSplines.size();
   }
   else{
       if(selected>=0 && selected<nbSplines){//single selection
           copiedBlur = splines[selected].getMeanBlur();
       }
   }
}

//paste the copied blur on the selected splines
void SplineSet::pasteBlur()
{
   if(copiedBlur!=-1){
       if(!selectedSplines.isEmpty()){//multiple selection
           for(int k=0;k<selectedSplines.size();k++)
           {
               splines[selectedSplines[k]].pasteBlur(copiedBlur);
           }
       }
       else{
           if(selected>=0 && selected<nbSplines){//single selection
               splines[selected].pasteBlur(copiedBlur);
           }
       }
   }
}

void SplineSet::sampleFromExample()
{
   if(selected>=0 && selected<nbSplines){//single selection
       splines[selected].sampleFromExample();
   }
}

/**********************************************************/

/*
load the curve set data from an XML file.
XML tag structure:
<curve_set>
<curve> global_len
nb_control_points
nb_left_colors
nb_right_colors
nb_blur_points
lifetime
<control_points_set>
<control_point> flagLine, x, y, globalID
</control_point>
....
</control_points_set>

<left_colors_set>
<left_color>
R,G,B, globalID
</left_color>
....
</left_colors_set>

<right_colors_set>
<right_color>
R,G,B, globalID
</right_color>
....
</right_colors_set>

<blur_points_set>
<best_scale>
value
</best_scale>
....
</blur_points_set>

</curve>
........
<curve>
....
</curve>
</curve_set>
*/
/**************************************************************************************************/
bool SplineSet::loadXML(QString fileName)
{

   //I make the supposition that the xml file is correctly filled-in.
   //I'm not checking that each node is where it's supposed to be.

   if (fileName.isEmpty()) return false;
   if (!QFile::exists (fileName)){
       cerr<<"[!]> The file \""<<fileName.data()<<"\" does not exist!"<<endl;
       return false;
   }

   // Open the specified file in Write-Only Mode
   QFile file (fileName);
   if (!file.open (QIODevice::ReadOnly)) return false;

   // Fill the QDomDocument
   QDomDocument doc ("CurveSetXML");
   if (!doc.setContent (&file)) {
       file.close();
       return false;
   }
   file.close();

   //clear the currently stored data
   reset();

   // Root node. Get the attributes ( image_width, image_height, nb_curves )
   const QDomElement root=doc.documentElement();
   imageWidth = ( root.hasAttribute("image_width") ) ? root.attribute("image_width").toInt():0;
   imageHeight = ( root.hasAttribute("image_height") ) ? root.attribute("image_height").toInt():0;
   //nbSplines = ( root.hasAttribute("nb_curves") ) ? root.attribute("nb_curves").toInt():0;

   //parse all curves
   for(QDomElement curve_node=root.firstChildElement("curve");!curve_node.isNull();curve_node=curve_node.nextSiblingElement())
   {
       //insertion mode (insert the curve in the set)
       insert();

       //each curve has:
       //Attributes: global_len, nb_control_points, nb_left_colors, nb_right_colors, nb_blur_points, lifetime
       //Children: a control_points_set, a left_colors_set, a right_colors_set and a blur_points_set

       //read attributes
      splines[selected].setLifetime(curve_node.attribute("lifetime").toInt());
       //curve.controlSize = curve_node.attribute("nb_control_points").toInt();
       //fill in the data
       //4 sets: control points, left color, right color, blur

       //CONTROL POINTS
       QDomElement ctrlPointsSet = curve_node.firstChildElement("control_points_set");
       if(!ctrlPointsSet.isNull())
       {
           for(QDomElement ctrlPoint = ctrlPointsSet.firstChildElement("control_point");!ctrlPoint.isNull();ctrlPoint=ctrlPoint.nextSiblingElement())
           {
               /*//flagLine
               CurveType t;
               string flag = ctrlPoint.attribute("flagLine").toStdString();
               if( flag.compare("line") == 0)
               t = LINE;
               else if( flag.compare("catmull") == 0)
               t = CURVE;
               else //end point
               t = EXTREM;
               curve.tags.push_back(t);*/
               //Position for a control point
               double x=0;
               double y=0;//position
               y = 1-(ctrlPoint.attribute("x").toDouble())/(0.5*(double)imageHeight);
               x = (ctrlPoint.attribute("y").toDouble())/(0.5*(double)imageWidth)-1;
               //add the control point to the current curve
               addPoint(x,y);
           }
       }
       //LEFT COLOR POINTS
       QDomElement leftColorsSet = curve_node.firstChildElement("left_colors_set");
       if(!leftColorsSet.isNull())
       {
           for(QDomElement leftColorPoint = leftColorsSet.firstChildElement("left_color");!leftColorPoint.isNull();leftColorPoint=leftColorPoint.nextSiblingElement())
           {
               QColor lC;//left color
               int lC_ID;
              lC.setRed((int)floor(leftColorPoint.attribute("B").toDouble()));
              lC.setGreen((int)floor(leftColorPoint.attribute("G").toDouble()));
              lC.setBlue((int)floor(leftColorPoint.attribute("R").toDouble()));
               lC_ID = (int)floor(leftColorPoint.attribute("globalID").toDouble());
               addLeftColorPoint(lC_ID, false, lC);
               //std::cerr<<"Left color nb. "<<k<<"is: (R = "<<lC.red()<<", G = "<<lC.green()<<", B = "<<lC.blue()<<")\n";
           }

       }
       //RIGHT COLOR POINTS
       QDomElement rightColorsSet = curve_node.firstChildElement("right_colors_set");
       if(!rightColorsSet.isNull())
       {
           for(QDomElement rightColorPoint = rightColorsSet.firstChildElement("right_color");!rightColorPoint.isNull();rightColorPoint=rightColorPoint.nextSiblingElement())
           {
               QColor rC;//right color
               int rC_ID;
              rC.setRed((int)floor(rightColorPoint.attribute("B").toDouble()));
              rC.setGreen((int)floor(rightColorPoint.attribute("G").toDouble()));
              rC.setBlue((int)floor(rightColorPoint.attribute("R").toDouble()));
               rC_ID = (int)floor(rightColorPoint.attribute("globalID").toDouble());
               addRightColorPoint(rC_ID, false, rC);
           }
       }
       //BLUR POINTS
       QDomElement blurPointsSet = curve_node.firstChildElement("blur_points_set");
       if(!blurPointsSet.isNull())
       {
           for(QDomElement blurPoint = blurPointsSet.firstChildElement("best_scale");!blurPoint.isNull();blurPoint=blurPoint.nextSiblingElement())
           {

               int bestScale = blurPoint.attribute("value").toInt();
               int s_ID = blurPoint.attribute("globalID").toInt();
               addBlurPoint(s_ID, false, bestScale);
           }
       }
   } //end curves

   //deselect all
   deselectAll();

   cout<<nbSplines<<endl;
   return true;
}

/**************************************************************************************************/
/*
save the curve set data to an XML file.
XML tag structure:
<curve_set>
<curve> global_len
nb_control_points
nb_left_colors
nb_right_colors
nb_blur_points
lifetime
<control_points_set>
<control_point> flagLine, x, y, globalID
</control_point>
....
</control_points_set>

<left_colors_set>
<left_color>
R,G,B, globalID
</left_color>
....
</left_colors_set>

<right_colors_set>
<right_color>
R,G,B, globalID
</right_color>
....
</right_colors_set>

<blur_points_set>
<best_scale>
value
</best_scale>
....
</blur_points_set>

</curve>
........
<curve>
....
</curve>
</curve_set>
*/
/**************************************************************************************************/
bool SplineSet::saveXML(QString fileName)
{
   //Create the root node: curve_set, with attributes ( image_width, image_height, nb_curves)
   QDomDocument doc( "CurveSetXML" );
   QDomElement root = doc.createElement( "curve_set" );
   root.setAttribute( "image_width", imageWidth);
   root.setAttribute( "image_height", imageHeight);
   root.setAttribute( "nb_curves", nbSplines);
   doc.appendChild( root );


   //for each curve, create a node, child of curve_set
   for(int i=0; i<nbSplines;i++)
   {
       //create the curve tag, with attributes (nb_control_points, lifetime)
       Catmull* curve = &(splines[i]);
       int nbCtrlPoints = curve->getNbPoint();
       int nbLeftColors = curve->getNbLeftColors();
       int nbRightColors = curve->getNbRightColors();
       int nbSigmas = curve->getNbBlurPoints();
       QDomElement curveTag = doc.createElement( "curve" );
       //curveTag.setAttribute( "global_len", curve->pixelChain.size());
       curveTag.setAttribute( "nb_control_points", nbCtrlPoints);
       curveTag.setAttribute( "nb_left_colors", nbLeftColors);
       curveTag.setAttribute( "nb_right_colors", nbRightColors);
       curveTag.setAttribute( "nb_blur_points", nbSigmas);
       curveTag.setAttribute( "lifetime", curve->getLifetime());

       QDomElement ctrlPointsSet = doc.createElement( "control_points_set" );
       QDomElement leftColorsSet = doc.createElement( "left_colors_set" );
       QDomElement rightColorsSet = doc.createElement( "right_colors_set" );
       QDomElement blurPointsSet = doc.createElement( "blur_points_set" );

       curveTag.appendChild( ctrlPointsSet );
       curveTag.appendChild( leftColorsSet );
       curveTag.appendChild( rightColorsSet );
       curveTag.appendChild( blurPointsSet );


       //add the control points, as children of the curve tag


       //GLOBAL ID TO CORRECT!
       for(int p=0; p<nbCtrlPoints;p++)
       {

           QDomElement ctrlPoint = doc.createElement( "control_point" );

           ///////////////////////////////////
           //position
           ///////////////////////////////////
           ctrlPoint.setAttribute( "y", 0.5*imageWidth*(curve->getCtrlPoint(p).getX()+1));
           ctrlPoint.setAttribute( "x", 0.5*imageHeight*(1-curve->getCtrlPoint(p).getY()));
           //add the control point data (flag, position) to the curve tag
           ctrlPointsSet.appendChild( ctrlPoint );
       }


       ///////////////////////////////////
       //ATTRIBUTES
       ///////////////////////////////////
       //left color
       for(int p=0; p<nbLeftColors;p++)
       {
           QColor attribVal;
           attribVal = curve->getLeftColor(p);
           QDomElement leftColor = doc.createElement( "left_color" );
           leftColor.setAttribute( "B", attribVal.red());
           leftColor.setAttribute( "G", attribVal.green());
           leftColor.setAttribute( "R", attribVal.blue());
           leftColor.setAttribute( "globalID", curve->getLeftColorID(p));
           leftColorsSet.appendChild( leftColor );
       }
       //right color
       for(int p=0; p<nbRightColors;p++)
       {
           QColor attribVal;
           attribVal = curve->getRightColor(p);
           QDomElement rightColor = doc.createElement( "right_color" );
           rightColor.setAttribute( "B", attribVal.red());
           rightColor.setAttribute( "G", attribVal.green());
           rightColor.setAttribute( "R", attribVal.blue());
           rightColor.setAttribute( "globalID", curve->getRightColorID(p));
           rightColorsSet.appendChild( rightColor );
       }


       ///////////////////////////////////
       //best scale
       for(int p=0; p<nbSigmas;p++)
       {
           int bestScaleVal = curve->getBlur(p);
           QDomElement bestScale = doc.createElement( "best_scale" );
           bestScale.setAttribute( "value", bestScaleVal);
           bestScale.setAttribute( "globalID", curve->getBlurID(p));
           blurPointsSet.appendChild( bestScale );
       }
       //add the curve tag to the curve_set
       root.appendChild( curveTag );
   }

   ///////////////////////////////////
   //save the XML structure in a file

   QFile file( fileName );
   if( !file.open( QIODevice::WriteOnly ) )
       return false;

   QTextStream ts(&file);
   ts << doc.toString();

   file.close();
   return true;
}

//*******************************************************************************
void SplineSet::setDragPoint(double x, double y, bool _draggingSpline)//initiate dragging a selected spline
{
   if(_draggingSpline && selected>=0 && selected<nbSplines)
       splines[selected].setDragPoint(x,y);
   //select multi
   else if(selectedSplines.size()>0)
   {
       for (int k=0;k<selectedSplines.size();k++)
           splines[selectedSplines[k]].setDragPoint(x,y);
   }
   draggingSpline  =_draggingSpline;
}
void SplineSet::dragSpline(double x, double y)//drag a selected spline
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].dragSpline(x,y);
   else if(selectedSplines.size()>0)
   {
       for (int k=0;k<selectedSplines.size();k++)
           splines[selectedSplines[k]].dragSpline(x,y);
       updateBoundingBox();
       updateConvexHull();
   }
}
bool SplineSet::isDraggingSpline()
{
   return draggingSpline;
}

//---------------//
//---- Snake ----//
//---------------//
void SplineSet::iterSnake()
{
   if(selected>=0 && selected<nbSplines)
       splines[selected].iterSnake();
}


//---------------------------------------------------------------//
//------    Vectorization of the drawing inputs    --------------//
//---------------------------------------------------------------//

void SplineSet::fitSpline(QVector<Point> drawFreeHandZoomed, int imageWidth, int imageHeight, double zoom, double centerX, double centerY)
{
   bool closedCurve = false;
   int len = drawFreeHandZoomed.size();
   int i;
   double x, y;

      bezierControlPoints.clear();
   polLinePoints.clear();
   BezierFit bez;
   /*for(i=0;i<len;i++)
   {
       drawFreeHandPath[i].y = floor((0.5*(double)imageHeight)*(1-drawFreeHandPath[i].y));
       drawFreeHandPath[i].x = floor((0.5*(double)imageWidth)*(1+drawFreeHandPath[i].x));
   }*/
   closedCurve = bez.fitBezier(drawFreeHandZoomed, bezierControlPoints, polLinePoints, zoom);



   //insert the spline
   insert();
   setClosedCurve(closedCurve);

   len = bezierControlPoints.size();
   for(int i=0;i<len;i++)
   {
       x = (bezierControlPoints[i].getX())/(0.5*(double)imageWidth)-1;
       y = 1-(bezierControlPoints[i].getY())/(0.5*(double)imageHeight);
       x/=zoom;
       y/=zoom;
       x-= centerX;
       y-= centerY;
       addPoint(x, y);
   }

   ////debug: add polyline points
   //    len = polLinePoints.size();

   //for(int i=0;i<len;i++)
   //{
   //    x = (polLinePoints[i].getX())/(0.5*(double)imageWidth)-1;
   //    y = 1-(polLinePoints[i].getY())/(0.5*(double)imageHeight);
   //    x/=zoom;
   //    y/=zoom;
   //    x-= centerX;
   //    y-= centerY;
   //    addPolLinePoint(x, y);
   //}

   //when drawn, the colors and blur are just at the beginning and at the end;
   len = bezierControlPoints.size();
   addLeftColorPoint((int)0);
   addRightColorPoint((int)0);
   addBlurPoint(0, true);

   addLeftColorPoint(((len-1)/3)*Catmull::NB_STEPS);
   addRightColorPoint(((len-1)/3)*Catmull::NB_STEPS);
   addBlurPoint(((len-1)/3)*Catmull::NB_STEPS, true);
   //we need to have all the points before setting the curve closed, because we use first and last point infos
   setClosedCurve(closedCurve);

   bezierControlPoints.clear();

   //if we need to sample the colors in the image
   if(!activateLeft){
       splines[selected].sampleColors(COLOR_LEFT);
   }
   if(!activateRight){
       splines[selected].sampleColors(COLOR_RIGHT);
   }
}

/*************************************************************************************/
int SplineSet::getNbRefPoint(){
	int i, sum=0;

	for(i=0;i<nbSplines;i++){
		sum+=splines[i].getNbConstraint();
	}
	return sum;
}

/*************************************************************************************/
Catmull SplineSet::getSplineCurb(int i){
	return splines[i];
}

/*************************************************************************************/
int SplineSet::getNbSplines(){
	return nbSplines;
}


