#ifndef SPLINESET_H
#define SPLINESET_H

#include <QVector>
#include "catmull.h"
#include "global_param.h"

class SplineSet  
{
public:
	static void interpolate(SplineSet* s1, SplineSet* s2, SplineSet* sResult, double t);//interpolate between s1 and s2: sResult = t*s2 + (1-t)*s1, 0<=t<=1
	//precondition: s1 and s2 have the same number of splines and control points
	bool loadXML(QString filename);
	bool saveXML(QString fileName);
	void draw();//draw the curves; if a curve is selected, draw the bounding box and control points
	void drawWithNameTags(ManipulationMode splineManip, ClickType clickType);//draw the curves with name tages attached
	void drawNormalX();//draw the curves with the X derivative color
	void drawNormalY();//draw the curves with the Y derivative color
	void drawColorConstraints(double zoom);//draw the curves with the color constraints on each side
	void drawOverlap(double zoom);//draw in the stencil buffer to detect overlapping areas
	void drawBlurAmount();//draw the amount of blur per pixel
	void drawControlPoints();//draw the spline control points of the selected curve
	void increaseBlur();//increase the blur of the selected spline
	void decreaseBlur();//decrease the blur of the selected spline
        //--- drawing ---//
	void setBlur(int s);//set the blur of the selected spline
	void setLeftColor(QColor C);//change the left color of the selected spline
	void setRightColor(QColor C);//change the right color of the selected spline
        void setActivateLeft(bool cl){activateLeft = cl;};//activate or deactivate left colors (if deactivated, the color is sampled in the reference image) 
        void setActivateRight(bool cr){activateRight = cr;};//activate or deactivate right colors (if deactivated, the color is sampled in the reference image) 
	void sampleFromExample();

	//change attributes globally
	void setGlobalBlur(int shift_sigma);
	//smooth the selected splines
	void smoothSplines();
        //activate or deactivate the stylization on the selected splines
        void stylizeSplines();
        //contrast of the selected splines
        void increaseContrast();
        void decreaseContrast();
	//change the color of the moving point for the selected spline
	void setControlPointLeftColor(QColor C,int id);
	void setControlPointRightColor(QColor C, int id);
	void setSigma(int sigma, int id);//set the sigma at control point having the id "id" in the selected spline
        //switch left and right colors of the selected splines
        void switchColors();

	QColor getControlPointLeftColor(int id);//get the left color of  moving point in the selected spline
	QColor getControlPointRightColor(int id);//get the right color of the  moving point in the selected spline
	int getSigma(int id);//get the sigma at control point having the id "id" in the selected spline


	QColor getLeftColor();//get the left color of the selected spline
	QColor getRightColor();//get the right color of the selected spline
	int getBlur();//get the blur of the selected spline

	//select one
	void setSelected(int id);
	int getSelected();
	//select multi
	bool areMultiSplinesSelected() {return !(selectedSplines.isEmpty());};
	void addRemoveToSelection(int id);
	void addToSelection(int id);
	void removeFromSelection(int id);
	void deselectAll();
	void deleteSelection();

	//copy paste
	void copySplines();
	void cutSplines();
	void pasteSplines(double x, double y, double center_x, double center_y, double zoom);
        void copyBlur();//copy the mean blur of the selected splines
        void pasteBlur();//paste the copied blur on the selected splines


	void setShowAttrib(bool _showAttrib) { showAttrib = _showAttrib;};

	void insert();
	void addSpline(Catmull C);
	void removeSpline(int id);
	void fitSpline(QVector<Point> drawFreeHandPath, int imageWidth, int imageHeight, double zoom, double centerX, double centerY);

	void setClosedCurve(bool _closedCurve);
	bool isClosedCurve();

	int getMovingPoint();//return the current moving point
	void addPoint(double x,double y);//add the (x,y) point in the control point list of the current spline, read the left or right color in the reference buffer if refL or refR is true
	//debug
	void addPolLinePoint(double x,double y);
	void insertPoint(double x,double y, int id, double t);//insert at the proper place the (x,y) point in the control point list of the current spline
        void split(double x,double y, int id, double t);
	void erasePoint(int id);//erase the control point at id pos from the list of the current spline
	void setMovingPoint(int id);//set the control point i to the be current moving point in the selected spline
	bool movePoint(double x,double y);//move the current moving point to (x,y) in the selected spline
	//return: true if a point is currently moving 
	void releaseMovingPoint();//release the current moving point

	void addLeftColorPoint(bool refL=false, double x=0,double y=0);
	void addRightColorPoint(bool refR=false, double x=0,double y=0);

	void addLeftColorPoint(int colorID, bool useDefault=true, QColor lC = QColor(0,0,0));//use the default color
	void addRightColorPoint(int colorID, bool useDefault=true, QColor rC = QColor(0,0,0));//use the default color
	void addBlurPoint(int blurID, bool useDefault, double sigmaBlur=0);//use the default blur


	void setDragPoint(double x, double y, bool _draggingSpline);//initiate dragging a selected spline
	void dragSpline(double x, double y);//drag a selected spline
	bool isDraggingSpline();

	//insert a control point at a position given by globalID in the attribType control points list
	// of the current spline
	void insertAttribPoint(int globalID, AttribType attribType);
	//erase the control point closest to globalID in the attribType control points list
	void eraseAttribPoint(int globalID, AttribType attribType);

	void init(double pSize, int wR, int hR, GLuint idR, QImage* img, QImage* grad);//Psize=pixel size, (wR,hR)=size of the reference buffer, idR=id of the reference buffer, img and grad are the exemple image and its gradient norm
	void reset();//clear the set of splines

	//transmit to the viewer the image width and height, after loadXML
	int getImageWidth() {return imageWidth;};
	int getImageHeight() {return imageHeight;};
	
	
	////////////////////////////////////////////////////////////////////////////////////////////////
	//return the total number of Color constraint point
	int getNbRefPoint();
	
	//return the number of spline
	int getNbSplines();
	
	//return the i_th catmull of the spline set
	Catmull getSplineCurb(int i);
	///////////////////////////////////////////////////////////////////////////////////////////////
	
	
	//snake
	void iterSnake();//iterate the snake of the selected spline

	SplineSet();
	~SplineSet();
private:
	QVector<Catmull> splines;//list of the splines
	int nbSplines;//number of splines in the set
	QVector<int> selectedSplines; //pointers towards the selected splines
	QVector<Catmull> copiedSplines; //pointers towards the copied splines
        int copiedBlur;
	int selected;//id of the currently selected spline
	//the bounding box left-up, right-down corners;
	Point leftUp;
	Point rightDown;
	void updateBoundingBox();
	void drawBoundingBox();
        //convex hull
        QVector<Point> convex_hull;
        void updateConvexHull();
        void drawConvexHull();
	
	//memebers used for fitting splines
	QVector<Point> bezierControlPoints;
	QVector<Point> polLinePoints;
	
	//MEMBERS

	QColor leftColor;
	QColor rightColor;
        bool activateLeft;
        bool activateRight;
	int sigma;//blur

	bool draggingSpline;
	bool showAttrib;

	//image width and height, for loadXML
	int imageWidth;
	int imageHeight;
};

#endif
